# stdlib
from typing import Any
from typing import Callable
from typing import Dict
from typing import List
from typing import Type
from typing import Union

# third party
from nacl.encoding import HexEncoder
from nacl.signing import SigningKey
from nacl.signing import VerifyKey
import requests

# syft absolute
import syft as sy

# relative
from ......grid import GridURL
from ......logger import error
from ......logger import info
from .....common.message import ImmediateSyftMessageWithReply
from .....common.message import ImmediateSyftMessageWithoutReply
from .....common.message import SignedImmediateSyftMessageWithReply
from .....common.message import SignedImmediateSyftMessageWithoutReply
from ....domain_interface import DomainInterface
from ....enums import AssociationRequestResponses
from ...exceptions import AuthorizationError
from ...exceptions import MissingRequestKeyError
from ...node_service.vpn.vpn_messages import VPNStatusMessageWithReply
from ...node_table.association_request import AssociationRequest
from ..auth import service_auth
from ..node_service import ImmediateNodeServiceWithReply
from ..node_service import ImmediateNodeServiceWithoutReply
from ..success_resp_message import ErrorResponseMessage
from ..success_resp_message import SuccessResponseMessage
from .association_request_messages import DeleteAssociationRequestMessage
from .association_request_messages import GetAssociationRequestMessage
from .association_request_messages import GetAssociationRequestResponse
from .association_request_messages import GetAssociationRequestsMessage
from .association_request_messages import GetAssociationRequestsResponse
from .association_request_messages import ReceiveAssociationRequestMessage
from .association_request_messages import RespondAssociationRequestMessage
from .association_request_messages import SendAssociationRequestMessage


def get_vpn_status_metadata(node: DomainInterface) -> Dict[str, Any]:
    vpn_status_msg = (
        VPNStatusMessageWithReply()
        .to(address=node.address, reply_to=node.address)
        .sign(signing_key=node.signing_key)
    )
    vpn_status = node.recv_immediate_msg_with_reply(msg=vpn_status_msg)
    vpn_status_message_contents = vpn_status.message
    status = vpn_status_message_contents.payload.kwargs  # type: ignore
    network_vpn_ip = status["host"]["ip"]
    node_name = status["host"]["hostname"]
    metadata = {
        "host_or_ip": str(network_vpn_ip),
        "node_id": str(node.target_id.id.no_dash),
        "node_name": str(node_name),
        "type": f"{str(type(node).__name__).lower()}",
    }
    return metadata


def check_if_is_vpn(host_or_ip: str) -> bool:
    VPN_IP_SUBNET = "100.64.0."
    return host_or_ip.startswith(VPN_IP_SUBNET)


# domain gets this message from a user and will try to send to the network
def send_association_request_msg(
    msg: SendAssociationRequestMessage,
    node: DomainInterface,
    verify_key: VerifyKey,
) -> Union[ErrorResponseMessage, SuccessResponseMessage]:
    # Check Key permissions
    info(
        f"Node {node} - send_association_request_msg: got SendAssociationRequestMessage. "
        f"Info: {msg.source} - {msg.target}"
    )
    allowed = node.users.can_manage_infrastructure(verify_key=verify_key)
    if allowed:
        user = node.users.get_user(verify_key=verify_key)
        info(
            f"Node {node} - send_association_request_msg: {node} got user performing the action. User: {user}"
        )

        # Build an association request to send to the target
        user_priv_key = SigningKey(
            node.users.get_user(verify_key).private_key.encode(), encoder=HexEncoder  # type: ignore
        )

        metadata = dict(msg.metadata)
        # get domain metadata to send to the network
        try:
            # TODO: refactor to not stuff our vpn_metadata into the normal metadata
            # because it gets blindly **splatted into the database
            vpn_metadata = get_vpn_status_metadata(node=node)
            metadata.update(vpn_metadata)
        except Exception as e:
            error(f"failed to get vpn status. {e}")

        metadata["node_name"] = (
            node.name if node.name else ""
        )  # tell the network what our name is

        try:
            # create a client to the target
            grid_url = GridURL.from_url(msg.target).with_path("/api/v1")
            target_client = sy.connect(url=str(grid_url), timeout=10)
        except requests.exceptions.ConnectTimeout:
            return ErrorResponseMessage(
                address=msg.reply_to,
                resp_msg="ConnectionTimeoutError: Node was not able to process your request in time.",
            )

        metadata["node_address"] = node.id.no_dash

        target_msg: SignedImmediateSyftMessageWithoutReply = (
            ReceiveAssociationRequestMessage(
                address=target_client.address,
                metadata=metadata,
                source=vpn_metadata["host_or_ip"],
                target=msg.target,
            ).sign(signing_key=user_priv_key)
        )

        # Send the message to the target
        info(
            f"Node {node} - send_association_request_msg: sending ReceiveAssociationRequestMessage."
        )
        try:
            # we need target
            target_client.send_immediate_msg_without_reply(msg=target_msg)
        except Exception as e:
            error(f"Failed to send ReceiveAssociationRequestMessage. {e}")
            error(f"Sending target message: {target_msg}")
            target_msg_args = {
                "address": target_client.address,
                "metadata": metadata,
                "source": vpn_metadata["host_or_ip"],
                "target": "msg.target",
            }
            error(f"Sending target message: {target_msg}")
            error(f"Sending target message: {target_msg_args}")
            raise e

        info(
            f"Node {node} - send_association_request_msg: received the answer from ReceiveAssociationRequestMessage."
        )

        # Create a new association request object
        info(
            f"Node {node} - send_association_request_msg: adding requests to the Database."
        )
        node.association_requests.create_association_request(
            node_name=target_client.name,  # type: ignore
            node_address=target_client.target_id.id.no_dash,  # type: ignore
            status=AssociationRequestResponses.PENDING,
            source=msg.source,
            target=msg.target,
        )
    else:  # If not authorized
        raise AuthorizationError("You're not allowed to create an Association Request!")

    info(f"Node: {node} received the answer from ReceiveAssociationRequestMessage.")

    return SuccessResponseMessage(
        address=msg.reply_to,
        resp_msg="Association request sent!",
    )


# network gets the above message first and then later the domain gets this message as well
def recv_association_request_msg(
    msg: ReceiveAssociationRequestMessage,
    node: DomainInterface,
    verify_key: VerifyKey,
) -> None:
    _previous_request = node.association_requests.contain(
        source=msg.source, target=msg.target
    )
    info(
        f"Node {node} - recv_association_request_msg: prev request exists {not _previous_request}."
    )

    # Create a new Association Request if the handshake value doesn't exist in the database
    if not _previous_request:
        # this side happens on the network
        info(
            f"Node {node} - recv_association_request_msg: creating a new association request."
        )

        if node.settings.DOMAIN_ASSOCIATION_REQUESTS_AUTOMATICALLY_ACCEPTED:
            status = AssociationRequestResponses.ACCEPT
        else:
            status = AssociationRequestResponses.PENDING

        node_address = msg.metadata["node_address"]

        node.association_requests.create_association_request(
            node_name=msg.metadata["node_name"],
            node_address=node_address,
            status=status,
            source=msg.source,
            target=msg.target,
        )
    else:
        # this side happens on the domain
        info(
            f"Node {node} - recv_association_request_msg: answering an existing association request."
        )
        node.association_requests.set(source=msg.source, target=msg.target, response=msg.response)  # type: ignore

    # get or create a new node and node_route which represents the opposing node which
    # is supplied in the metadata
    try:
        node_id = node.node.create_or_get_node(  # type: ignore
            node_uid=msg.metadata["node_id"], node_name=msg.metadata["node_name"]
        )
        is_vpn = check_if_is_vpn(host_or_ip=msg.metadata["host_or_ip"])
        node.node_route.update_route_for_node(  # type: ignore
            node_id=node_id, host_or_ip=msg.metadata["host_or_ip"], is_vpn=is_vpn
        )
    except Exception as e:
        error(f"Failed to save the node and node_route rows. {e}")


# network owner user approves the request and sends this to the network
def respond_association_request_msg(
    msg: RespondAssociationRequestMessage,
    node: DomainInterface,
    verify_key: VerifyKey,
) -> Union[ErrorResponseMessage, SuccessResponseMessage]:
    # Check if handshake/address/value fields are empty
    missing_paramaters = not msg.target or not msg.response
    if missing_paramaters:
        raise MissingRequestKeyError(
            message="Invalid request payload, empty fields (target/handshake/value)!"
        )
    # Check Key permissions
    allowed = node.users.can_manage_infrastructure(verify_key=verify_key)
    resp_msg = "Association request replied!"

    info(
        f"Node {node} - respond_association_request_msg: user can approve/deny association requests."
    )
    if allowed:
        # Set the status of the Association Request according to the "value" field received

        node.association_requests.set(source=msg.source, target=msg.target, response=msg.response)  # type: ignore
        user_priv_key = SigningKey(
            node.users.get_user(verify_key).private_key.encode(), encoder=HexEncoder  # type: ignore
        )

        metadata = {}
        # get network metadata to send back to domain
        try:
            metadata = get_vpn_status_metadata(node=node)
        except Exception as e:
            error(f"Failed to get vpn status. {e}")

        try:
            # create a client to the source
            grid_url = GridURL.from_url(msg.source).with_path("/api/v1")
            source_client = sy.connect(url=str(grid_url), timeout=10)
        except requests.exceptions.ConnectTimeout:
            return ErrorResponseMessage(
                address=msg.reply_to,
                resp_msg="Timeout error node was not able to process your request in time.",
            )

        try:
            metadata["node_address"] = node.id.no_dash  # type:ignore
            node_msg: SignedImmediateSyftMessageWithReply = (
                ReceiveAssociationRequestMessage(
                    address=source_client.address,
                    response=msg.response,
                    metadata=metadata,
                    source=msg.source,
                    target=msg.target,
                ).sign(signing_key=user_priv_key)
            )

            info(
                f"Node {node} - respond_association_request_msg: sending ReceiveAssociationRequestMessage."
            )

            source_client.send_immediate_msg_without_reply(msg=node_msg)

            info(
                f"Node {node} - respond_association_request_msg: ReceiveAssociationRequestMessage got back."
            )
        except Exception as e:
            error(f"Failed to send ReceiveAssociationRequestMessage to the domain. {e}")

    else:  # If not allowed
        raise AuthorizationError("You're not allowed to create an Association Request!")

    return SuccessResponseMessage(
        address=msg.reply_to,
        resp_msg=resp_msg,
    )


def get_association_request_msg(
    msg: GetAssociationRequestMessage,
    node: DomainInterface,
    verify_key: VerifyKey,
) -> GetAssociationRequestResponse:
    # Check Key Permissions
    allowed = node.users.can_manage_infrastructure(verify_key=verify_key)
    # If allowed
    if allowed:
        association_request: AssociationRequest = node.association_requests.first(
            id=msg.association_id
        )
    else:  # Otherwise
        raise AuthorizationError(
            "You're not allowed to get Association Request information!"
        )

    return GetAssociationRequestResponse(
        address=msg.reply_to,
        content=association_request.get_metadata(),
        source=association_request.source,
        target=association_request.target,
    )


def get_all_association_request_msg(
    msg: GetAssociationRequestsMessage,
    node: DomainInterface,
    verify_key: VerifyKey,
) -> GetAssociationRequestsResponse:
    allowed = node.users.can_manage_infrastructure(verify_key=verify_key)

    # If allowed
    if allowed:
        association_requests = node.association_requests.all()

        association_requests_json = [
            association_request.get_metadata()
            for association_request in association_requests
        ]
    else:  # Otherwise
        raise AuthorizationError(
            "You're not allowed to get Association Request information!"
        )

    return GetAssociationRequestsResponse(
        address=msg.reply_to,
        content=association_requests_json,
    )


def del_association_request_msg(
    msg: DeleteAssociationRequestMessage,
    node: DomainInterface,
    verify_key: VerifyKey,
) -> SuccessResponseMessage:
    # Check Key permissions
    allowed = node.users.can_manage_infrastructure(
        verify_key=verify_key
    ) and node.association_requests.contain(id=msg.association_id)

    # If allowed
    if allowed:
        node.association_requests.delete(id=msg.association_id)
    else:  # Otherwise
        raise AuthorizationError(
            "You're not allowed to delete this Association Request information!"
        )

    return SuccessResponseMessage(
        address=msg.reply_to,
        resp_msg="Association request deleted!",
    )


class AssociationRequestService(ImmediateNodeServiceWithReply):

    msg_handler_map: Dict[type, Callable] = {
        SendAssociationRequestMessage: send_association_request_msg,
        GetAssociationRequestMessage: get_association_request_msg,
        GetAssociationRequestsMessage: get_all_association_request_msg,
        DeleteAssociationRequestMessage: del_association_request_msg,
        RespondAssociationRequestMessage: respond_association_request_msg,
    }

    @staticmethod
    @service_auth(guests_welcome=True)
    def process(
        node: DomainInterface,
        msg: Union[
            SendAssociationRequestMessage,
            GetAssociationRequestMessage,
            DeleteAssociationRequestMessage,
        ],
        verify_key: VerifyKey,
    ) -> Union[
        SuccessResponseMessage,
        ErrorResponseMessage,
        GetAssociationRequestsResponse,
        GetAssociationRequestResponse,
    ]:
        return AssociationRequestService.msg_handler_map[type(msg)](
            msg=msg, node=node, verify_key=verify_key
        )

    @staticmethod
    def message_handler_types() -> List[Type[ImmediateSyftMessageWithReply]]:
        return [
            SendAssociationRequestMessage,
            GetAssociationRequestMessage,
            GetAssociationRequestsMessage,
            DeleteAssociationRequestMessage,
            RespondAssociationRequestMessage,
        ]


class AssociationRequestWithoutReplyService(ImmediateNodeServiceWithoutReply):

    msg_handler_map: Dict[type, Callable] = {
        ReceiveAssociationRequestMessage: recv_association_request_msg,
    }

    @staticmethod
    @service_auth(guests_welcome=True)
    def process(
        node: DomainInterface,
        msg: ReceiveAssociationRequestMessage,
        verify_key: VerifyKey,
    ) -> None:
        return AssociationRequestWithoutReplyService.msg_handler_map[type(msg)](
            msg=msg, node=node, verify_key=verify_key
        )

    @staticmethod
    def message_handler_types() -> List[Type[ImmediateSyftMessageWithoutReply]]:
        return [
            ReceiveAssociationRequestMessage,
        ]



# stdlib
import functools
from typing import Any
from typing import Dict
from typing import List
from typing import Optional

# third party
from google.protobuf.reflection import GeneratedProtocolMessageType
from nacl.signing import VerifyKey

# syft absolute
import syft as sy

# relative
# from . import context
from ..... import lib
from .....logger import traceback_and_raise
from .....logger import warning
from .....proto.core.node.common.action.run_class_method_pb2 import (
    RunClassMethodAction as RunClassMethodAction_PB,
)
from .....util import inherit_tags
from ....common.serde.serializable import serializable
from ....common.uid import UID
from ....io.address import Address
from ....store.proxy_dataset import ProxyDataset
from ....store.storeable_object import StorableObject
from ...abstract.node import AbstractNode
from ..util import check_send_to_blob_storage
from ..util import upload_result_to_s3
from .common import ImmediateActionWithoutReply
from .greenlets_switch import retrieve_object


@serializable()
class RunClassMethodAction(ImmediateActionWithoutReply):
    """
    When executing a RunClassMethodAction, a :class:`Node` will run a method defined
    by the action's path attribute on the object pointed at by _self and keep the returned
    value in its store.

    Attributes:
         path: the dotted path to the method to call
         _self: a pointer to the object which the method should be applied to.
         args: args to pass to the function. They should be pointers to objects
            located on the :class:`Node` that will execute the action.
         kwargs: kwargs to pass to the function. They should be pointers to objects
            located on the :class:`Node` that will execute the action.
    """

    def __init__(
        self,
        path: str,
        _self: Any,
        args: List[Any],
        kwargs: Dict[Any, Any],
        id_at_location: UID,
        address: Address,
        msg_id: Optional[UID] = None,
        is_static: Optional[bool] = False,
    ):
        self.path = path
        self._self = _self
        self.args = args
        self.kwargs = kwargs
        self.id_at_location = id_at_location
        self.is_static = is_static
        # logging needs .path to exist before calling
        # this which is why i've put this super().__init__ down here
        super().__init__(address=address, msg_id=msg_id)

    @staticmethod
    def intersect_keys(
        left: Dict[VerifyKey, Optional[UID]], right: Dict[VerifyKey, Optional[UID]]
    ) -> Dict[VerifyKey, Optional[UID]]:
        # get the intersection of the dict keys, the value is the request_id
        # if the request_id is different for some reason we still want to keep it,
        # so only intersect the keys and then copy those over from the main dict
        # into a new one
        intersection = set(left.keys()).intersection(right.keys())
        # left and right have the same keys
        return {k: left[k] for k in intersection}

    @property
    def pprint(self) -> str:
        return f"RunClassMethodAction({self.path})"

    def __repr__(self) -> str:
        method_name = self.path.split(".")[-1]
        self_name = str(self._self.__class__.__name__)
        arg_names = ",".join([a.__class__.__name__ for a in self.args])
        kwargs_names = ",".join(
            [f"{k}={v.__class__.__name__}" for k, v in self.kwargs.items()]
        )
        return f"RunClassMethodAction {self_name}.{method_name}({arg_names}, {kwargs_names})"

    def execute_action(self, node: AbstractNode, verify_key: VerifyKey) -> None:
        method = node.lib_ast(self.path)

        mutating_internal = False
        if (
            self.path.startswith("torch.Tensor")
            and self.path.endswith("_")
            and not self.path.endswith("__call__")
        ):
            mutating_internal = True
        elif not self.path.startswith("torch.Tensor") and self.path.endswith(
            "__call__"
        ):
            mutating_internal = True

        resolved_self = None
        is_proxy = False  # we need to know if its a proxy object when we save mutations
        if not self.is_static:
            resolved_self = retrieve_object(
                node=node,
                id_at_location=self._self.id_at_location,
                path=self.path,
                proxy_only=True,
            )
            # check if its proxy and resolve
            if isinstance(resolved_self.data, ProxyDataset):
                is_proxy = True
                resolved_self.data = resolved_self.data.get_s3_data(
                    settings=node.settings
                )
            result_read_permissions = resolved_self.read_permissions  # type: ignore
            result_write_permissions = resolved_self.write_permissions  # type: ignore
        else:
            result_read_permissions = {}
            result_write_permissions = {}

        resolved_args = list()
        tag_args = []
        for arg in self.args:
            r_arg = retrieve_object(node, arg.id_at_location, self.path)
            result_read_permissions = self.intersect_keys(
                result_read_permissions, r_arg.read_permissions  # type: ignore
            )
            resolved_args.append(r_arg.data)  # type: ignore
            tag_args.append(r_arg)

        resolved_kwargs = {}
        tag_kwargs = {}
        for arg_name, arg in self.kwargs.items():
            r_arg = retrieve_object(node, arg.id_at_location, self.path)
            result_read_permissions = self.intersect_keys(
                result_read_permissions, r_arg.read_permissions  # type: ignore
            )
            resolved_kwargs[arg_name] = r_arg.data  # type: ignore
            tag_kwargs[arg_name] = r_arg

        (
            upcasted_args,
            upcasted_kwargs,
        ) = lib.python.util.upcast_args_and_kwargs(resolved_args, resolved_kwargs)

        resolved_self_previous_bytes: Optional[bytes] = None
        if self.is_static:
            result = method(*upcasted_args, **upcasted_kwargs)
        else:
            if resolved_self is None:
                traceback_and_raise(
                    ValueError(f"Method {method} called, but self is None.")
                )

            resolved_self_previous_bytes = sy.serialize(
                resolved_self.data, to_bytes=True
            )
            method_name = self.path.split(".")[-1]

            target_method = getattr(resolved_self.data, method_name, None)

            if id(target_method) != id(method):
                warning(
                    f"Method {method_name} overwritten on object {resolved_self.data}"
                )
                method = target_method
            else:
                method = functools.partial(method, resolved_self.data)

            result = method(*upcasted_args, **upcasted_kwargs)

        # TODO: add numpy support https://github.com/OpenMined/PySyft/issues/5164
        if "numpy." in str(type(result)):
            if "float" in type(result).__name__:
                result = float(result)
            if "int" in type(result).__name__:
                result = int(result)
            if "bool" in type(result).__name__:
                result = bool(result)

        if lib.python.primitive_factory.isprimitive(value=result):
            # Wrap in a SyPrimitive
            result = lib.python.primitive_factory.PrimitiveFactory.generate_primitive(
                value=result, id=self.id_at_location
            )
        else:
            # TODO: overload all methods to incorporate this automatically
            if hasattr(result, "id"):
                try:
                    if hasattr(result, "_id"):
                        # set the underlying id
                        result._id = self.id_at_location
                    else:
                        result.id = self.id_at_location

                    if result.id != self.id_at_location:
                        raise AttributeError("IDs don't match")
                except AttributeError as e:
                    err = f"Unable to set id on result {type(result)}. {e}"
                    traceback_and_raise(Exception(err))

        # TODO: Find a better way to do this, a serialization is compute intesive.
        # check if resolved_self has changed and if so mark as mutating_internal
        # this prevents someone from mutating an object they own with something they
        # do not own and the read_permissions not flowing backwards

        if (
            resolved_self_previous_bytes is not None
            and resolved_self is not None
            and resolved_self_previous_bytes
            != sy.serialize(resolved_self.data, to_bytes=True)
        ):
            mutating_internal = True

        if verify_key not in result_write_permissions:
            # User does not have permission write permissions to this pointer.
            # Therefore object mutation is not allowed.
            mutating_internal = False
            # TODO: Need to clarify with Madhava/Andrew if it should be allowed to
            # create the result pointer and store it in the database.
            # traceback_and_raise(
            #     Exception("You don't have permissions to perform the write operation.")
            # )

        if mutating_internal:
            if isinstance(resolved_self, StorableObject):
                resolved_self.read_permissions = result_read_permissions
                resolved_self.write_permissions = result_write_permissions

        # in memory lookup for publish_service.py:40
        # context.OBJ_CACHE[str(self.id_at_location.no_dash)] = result

        # TODO: Upload object to seaweed store, instead of storing in redis
        # create a proxy object class and store it here.
        if check_send_to_blob_storage(
            obj=result,
            use_blob_storage=getattr(node.settings, "USE_BLOB_STORAGE", False),
        ):
            result = upload_result_to_s3(
                asset_name=self.id_at_location.no_dash,
                dataset_name="",
                domain_id=node.id,
                data=result,
                settings=node.settings,
            )

        if not isinstance(result, StorableObject):
            result = StorableObject(
                id=self.id_at_location,
                data=result,
                read_permissions=result_read_permissions,
                write_permissions=result_write_permissions,
            )

        inherit_tags(
            attr_path_and_name=self.path,
            result=result,
            self_obj=resolved_self,
            args=tag_args,
            kwargs=tag_kwargs,
        )

        # if we have mutated resolved_self we need to save it back since the store
        # might be in SQL and not in memory where the update is automatic
        # but if the method was static then we might not have a _self
        if resolved_self is not None and mutating_internal:
            # write the original resolved_self back to _self.id_at_location
            if is_proxy:
                resolved_self_proxy = upload_result_to_s3(
                    asset_name=self._self.id_at_location.no_dash,
                    dataset_name="",
                    domain_id=node.id,
                    data=resolved_self.data,
                    settings=node.settings,
                )
                resolved_self.data = resolved_self_proxy
            node.store[self._self.id_at_location] = resolved_self  # type: ignore

        node.store[self.id_at_location] = result

    def _object2proto(self) -> RunClassMethodAction_PB:
        """Returns a protobuf serialization of self.

        As a requirement of all objects which inherit from Serializable,
        this method transforms the current object into the corresponding
        Protobuf object so that it can be further serialized.

        :return: returns a protobuf object
        :rtype: RunClassMethodAction_PB

        .. note::
            This method is purely an internal method. Please use sy.serialize(object) or one of
            the other public serialization methods if you wish to serialize an
            object.
        """

        return RunClassMethodAction_PB(
            path=self.path,
            _self=sy.serialize(self._self, to_bytes=True),
            args=list(map(lambda x: sy.serialize(x, to_bytes=True), self.args)),
            kwargs={k: sy.serialize(v, to_bytes=True) for k, v in self.kwargs.items()},
            id_at_location=sy.serialize(self.id_at_location),
            address=sy.serialize(self.address),
            msg_id=sy.serialize(self.id),
        )

    @staticmethod
    def _proto2object(proto: RunClassMethodAction_PB) -> "RunClassMethodAction":
        """Creates a ObjectWithID from a protobuf

        As a requirement of all objects which inherit from Serializable,
        this method transforms a protobuf object into an instance of this class.

        :return: returns an instance of RunClassMethodAction
        :rtype: RunClassMethodAction

        .. note::
            This method is purely an internal method. Please use syft.deserialize()
            if you wish to deserialize an object.
        """

        return RunClassMethodAction(
            path=proto.path,
            _self=sy.deserialize(blob=proto._self, from_bytes=True),
            args=list(
                map(lambda x: sy.deserialize(blob=x, from_bytes=True), proto.args)
            ),
            kwargs={
                k: sy.deserialize(blob=v, from_bytes=True)
                for k, v in proto.kwargs.items()
            },
            id_at_location=sy.deserialize(blob=proto.id_at_location),
            address=sy.deserialize(blob=proto.address),
            msg_id=sy.deserialize(blob=proto.msg_id),
        )

    @staticmethod
    def get_protobuf_schema() -> GeneratedProtocolMessageType:
        """Return the type of protobuf object which stores a class of this type

        As a part of serialization and deserialization, we need the ability to
        lookup the protobuf object type directly from the object type. This
        static method allows us to do this.

        Importantly, this method is also used to create the reverse lookup ability within
        the metaclass of Serializable. In the metaclass, it calls this method and then
        it takes whatever type is returned from this method and adds an attribute to it
        with the type of this class attached to it. See the MetaSerializable class for details.

        :return: the type of protobuf object which corresponds to this class.
        :rtype: GeneratedProtocolMessageType

        """

        return RunClassMethodAction_PB

    def remap_input(self, current_input: Any, new_input: Any) -> None:
        """Redefines some of the arguments, and possibly the _self of the function"""
        if self._self.id_at_location == current_input.id_at_location:
            self._self = new_input

        for i, arg in enumerate(self.args):
            if arg.id_at_location == current_input.id_at_location:
                self.args[i] = new_input

        for k, v in self.kwargs.items():
            if v.id_at_location == current_input.id_at_location:
                self.kwargs[k] = new_input



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: proto/core/tensor/tensor.proto
"""Generated protocol buffer code."""
# third party
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


# syft absolute
from syft.proto.lib.numpy import array_pb2 as proto_dot_lib_dot_numpy_dot_array__pb2

DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\x1eproto/core/tensor/tensor.proto\x12\x10syft.core.tensor\x1a\x1bproto/lib/numpy/array.proto"\x9d\x01\n\x06Tensor\x12\x10\n\x08obj_type\x18\x02 \x01(\t\x12\x13\n\x0buse_tensors\x18\x04 \x01(\x08\x12*\n\x06\x61rrays\x18\x05 \x03(\x0b\x32\x1a.syft.lib.numpy.NumpyProto\x12)\n\x07tensors\x18\x06 \x03(\x0b\x32\x18.syft.core.tensor.Tensor\x12\x15\n\rrequires_grad\x18\x07 \x01(\x08\x62\x06proto3'
)

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(
    DESCRIPTOR, "proto.core.tensor.tensor_pb2", globals()
)
if _descriptor._USE_C_DESCRIPTORS == False:

    DESCRIPTOR._options = None
    _TENSOR._serialized_start = 82
    _TENSOR._serialized_end = 239
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: proto/lib/torch/tensor.proto
"""Generated protocol buffer code."""
# third party
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


# syft absolute
from syft.proto.lib.torch import device_pb2 as proto_dot_lib_dot_torch_dot_device__pb2

DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\x1cproto/lib/torch/tensor.proto\x12\x0esyft.lib.torch\x1a\x1cproto/lib/torch/device.proto"\xe1\x02\n\x0fProtobufContent\x12\r\n\x05shape\x18\x01 \x03(\x03\x12\x16\n\x0e\x63ontents_uint8\x18\x10 \x03(\r\x12\x15\n\rcontents_int8\x18\x11 \x03(\x05\x12\x16\n\x0e\x63ontents_int16\x18\x12 \x03(\x05\x12\x16\n\x0e\x63ontents_int32\x18\x13 \x03(\x05\x12\x16\n\x0e\x63ontents_int64\x18\x14 \x03(\x03\x12\x18\n\x10\x63ontents_float16\x18\x15 \x03(\x02\x12\x18\n\x10\x63ontents_float32\x18\x16 \x03(\x02\x12\x18\n\x10\x63ontents_float64\x18\x17 \x03(\x01\x12\x15\n\rcontents_bool\x18\x18 \x03(\x08\x12\x16\n\x0e\x63ontents_qint8\x18\x19 \x03(\x11\x12\x17\n\x0f\x63ontents_quint8\x18\x1a \x03(\r\x12\x17\n\x0f\x63ontents_qint32\x18\x1b \x03(\x11\x12\x19\n\x11\x63ontents_bfloat16\x18\x1c \x03(\x02"\x88\x01\n\nTensorData\x12\x14\n\x0cis_quantized\x18\x01 \x01(\x08\x12\r\n\x05scale\x18\x02 \x01(\x01\x12\x12\n\nzero_point\x18\x03 \x01(\x05\x12\x14\n\nproto_data\x18\x04 \x01(\x0cH\x00\x12\x14\n\narrow_data\x18\x05 \x01(\x0cH\x00\x12\r\n\x05\x64type\x18\x06 \x01(\tB\x06\n\x04\x64\x61ta"j\n\x0bTensorProto\x12\x0e\n\x06tensor\x18\x01 \x01(\x0c\x12\x15\n\rrequires_grad\x18\x02 \x01(\x08\x12\x0c\n\x04grad\x18\x03 \x01(\x0c\x12&\n\x06\x64\x65vice\x18\x04 \x01(\x0b\x32\x16.syft.lib.torch.Deviceb\x06proto3'
)

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(
    DESCRIPTOR, "proto.lib.torch.tensor_pb2", globals()
)
if _descriptor._USE_C_DESCRIPTORS == False:

    DESCRIPTOR._options = None
    _PROTOBUFCONTENT._serialized_start = 79
    _PROTOBUFCONTENT._serialized_end = 432
    _TENSORDATA._serialized_start = 435
    _TENSORDATA._serialized_end = 571
    _TENSORPROTO._serialized_start = 573
    _TENSORPROTO._serialized_end = 679
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: proto/grid/messages/transfer_messages.proto
"""Generated protocol buffer code."""
# third party
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


# syft absolute
from syft.proto.core.common import (
    common_object_pb2 as proto_dot_core_dot_common_dot_common__object__pb2,
)
from syft.proto.core.io import address_pb2 as proto_dot_core_dot_io_dot_address__pb2

DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n+proto/grid/messages/transfer_messages.proto\x12\x12syft.grid.messages\x1a%proto/core/common/common_object.proto\x1a\x1bproto/core/io/address.proto"\x9c\x01\n\x11LoadObjectMessage\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12&\n\x07\x61\x64\x64ress\x18\x02 \x01(\x0b\x32\x15.syft.core.io.Address\x12\x0f\n\x07\x63ontent\x18\x03 \x01(\t\x12\'\n\x08reply_to\x18\x04 \x01(\x0b\x32\x15.syft.core.io.Address"\x89\x01\n\x12LoadObjectResponse\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12\x13\n\x0bstatus_code\x18\x02 \x01(\x05\x12\x0f\n\x07\x63ontent\x18\x03 \x01(\t\x12&\n\x07\x61\x64\x64ress\x18\x04 \x01(\x0b\x32\x15.syft.core.io.Address"s\n\x11SaveObjectMessage\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12&\n\x07\x61\x64\x64ress\x18\x02 \x01(\x0b\x32\x15.syft.core.io.Address\x12\x0f\n\x07\x63ontent\x18\x03 \x01(\t"\x89\x01\n\x12SaveObjectResponse\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12\x13\n\x0bstatus_code\x18\x02 \x01(\x05\x12\x0f\n\x07\x63ontent\x18\x03 \x01(\t\x12&\n\x07\x61\x64\x64ress\x18\x04 \x01(\x0b\x32\x15.syft.core.io.Addressb\x06proto3'
)

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(
    DESCRIPTOR, "proto.grid.messages.transfer_messages_pb2", globals()
)
if _descriptor._USE_C_DESCRIPTORS == False:

    DESCRIPTOR._options = None
    _LOADOBJECTMESSAGE._serialized_start = 136
    _LOADOBJECTMESSAGE._serialized_end = 292
    _LOADOBJECTRESPONSE._serialized_start = 295
    _LOADOBJECTRESPONSE._serialized_end = 432
    _SAVEOBJECTMESSAGE._serialized_start = 434
    _SAVEOBJECTMESSAGE._serialized_end = 549
    _SAVEOBJECTRESPONSE._serialized_start = 552
    _SAVEOBJECTRESPONSE._serialized_end = 689
# @@protoc_insertion_point(module_scope)



# future
from __future__ import annotations

# stdlib
from typing import Union

# third party
from nacl.encoding import HexEncoder
from nacl.signing import SigningKey

# relative
from ...core.common.uid import UID
from ...core.io.address import Address
from ...core.io.location import SpecificLocation
from ...core.io.route import SoloRoute
from ...core.node.common import AbstractNodeClient
from ...core.node.common.node_service.ping.ping_messages import PingMessageWithReply
from ...core.node.domain_client import DomainClient
from ...logger import error
from .grid_connection import GridHTTPConnection


class ProxyClient(DomainClient):
    def post_init(self) -> None:
        # use post_init so we don't need to re-implement an empty regular init
        super().post_init()
        self.logged_in = False

    def __repr__(self) -> str:
        if self.logged_in:
            return super().__repr__()
        else:
            return (
                f"(This is a logged out ProxyClient() object for a domain called '{self.name}'."
                f" Please call .login(email, password) to get a full client you can use for stuff.)"
            )

    @property
    def ping(self) -> bool:
        try:
            # Build Syft Message
            msg = (
                PingMessageWithReply(kwargs={"host_or_ip": "asdf"})
                .to(address=self.address, reply_to=self.address)
                .sign(signing_key=self.signing_key)
            )
            self.send_immediate_msg_with_reply(msg, timeout=1)
            return True
        except Exception:
            return False

    @staticmethod
    def create(
        proxy_node_client: AbstractNodeClient,
        remote_domain: Union[str, UID, Address],
        domain_name: str,
    ) -> ProxyClient:
        domain_address = remote_domain
        try:
            if isinstance(domain_address, str):
                domain_address = UID.from_string(value=domain_address)
        except Exception as e:
            error(f"Failed to convert remote_domain str to UID. {e}")

        try:
            if isinstance(domain_address, UID):
                spec_location = SpecificLocation(domain_address)
                domain_address = Address(name=domain_name, domain=spec_location)
        except Exception as e:
            error(f"Failed to convert remote_domain UID to Address. {e}")

        if not isinstance(domain_address, Address):
            raise Exception(
                f"Failed to convert remote_domain {domain_address} to Address"
            )

        # the DomainRequestAPI is only on Network Clients so here we can re-use
        # the existing client route which will be pointing to the current Network
        base_url = proxy_node_client.routes[0].connection.base_url  # type: ignore
        conn = GridHTTPConnection(url=base_url)

        # this will initially be a guest connection so we should generate a new
        # unique key
        _user_key = SigningKey.generate()

        # Create a new Solo Route using the selected connection type
        route = SoloRoute(destination=spec_location, connection=conn)

        # here we mix all of the above together into a ProxyClient which is like a
        # normal client, except the route and the address are for different nodes
        proxy_client = ProxyClient(
            name=domain_name,
            routes=[route],
            signing_key=_user_key,
            domain=domain_address.domain,  # type: ignore
        )

        return proxy_client

    def login(self, email: str, password: str) -> ProxyClient:
        result = self.users.login(email=email, password=password)
        try:
            if result["status"] == "ok":
                _user_key = SigningKey(
                    result["data"]["key"].encode(), encoder=HexEncoder
                )
                self.signing_key = _user_key
                self.verify_key = self.signing_key.verify_key
                print(f"Logged in to {self.name} as {email}")
                self.logged_in = True
            else:
                print(f"Failed to login as {email}")
        except Exception as e:
            print(f"Failed to decode private key. {e}")
        return self

    def logout(self) -> ProxyClient:
        print(f"Logged out of account. You are now a Guest on {self.name}")
        self.signing_key = SigningKey.generate()
        self.verify_key = self.signing_key.verify_key
        self.logged_in = False
        return self



# stdlib
from typing import List
from typing import Optional

# third party
from nacl.signing import VerifyKey

# relative
from ......logger import traceback_and_raise
from ....abstract.node import AbstractNode
from ..node_service import ImmediateNodeServiceWithReply
from .request_answer_messages import RequestAnswerMessage
from .request_answer_messages import RequestAnswerResponse


class RequestAnswerService(ImmediateNodeServiceWithReply):
    @staticmethod
    def message_handler_types() -> List[type]:
        return [RequestAnswerMessage]

    @staticmethod
    def process(
        node: AbstractNode,
        msg: RequestAnswerMessage,
        verify_key: Optional[VerifyKey] = None,
    ) -> RequestAnswerResponse:
        if verify_key is None:
            traceback_and_raise(
                ValueError(
                    "Can't process Request service without a given " "verification key"
                )
            )

        status = node.get_request_status(message_request_id=msg.request_id)  # type: ignore
        address = msg.reply_to
        return RequestAnswerResponse(
            request_id=msg.request_id, address=address, status=status
        )



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: proto/grid/messages/setup_messages.proto
"""Generated protocol buffer code."""
# third party
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


# syft absolute
from syft.proto.core.common import (
    common_object_pb2 as proto_dot_core_dot_common_dot_common__object__pb2,
)
from syft.proto.core.io import address_pb2 as proto_dot_core_dot_io_dot_address__pb2

DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n(proto/grid/messages/setup_messages.proto\x12\x12syft.grid.messages\x1a%proto/core/common/common_object.proto\x1a\x1bproto/core/io/address.proto"\xfc\x01\n\x19\x43reateInitialSetUpMessage\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12&\n\x07\x61\x64\x64ress\x18\x02 \x01(\x0b\x32\x15.syft.core.io.Address\x12\x0c\n\x04name\x18\x03 \x01(\t\x12\r\n\x05\x65mail\x18\x04 \x01(\t\x12\x10\n\x08password\x18\x05 \x01(\t\x12\x13\n\x0b\x64omain_name\x18\x06 \x01(\t\x12\x0e\n\x06\x62udget\x18\x07 \x01(\x01\x12\'\n\x08reply_to\x18\x08 \x01(\x0b\x32\x15.syft.core.io.Address\x12\x13\n\x0bsigning_key\x18\t \x01(\x0c"\x89\x01\n\x0fGetSetUpMessage\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12&\n\x07\x61\x64\x64ress\x18\x02 \x01(\x0b\x32\x15.syft.core.io.Address\x12\'\n\x08reply_to\x18\x03 \x01(\x0b\x32\x15.syft.core.io.Address"\x87\x01\n\x10GetSetUpResponse\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12\x13\n\x0bstatus_code\x18\x02 \x01(\x05\x12\x0f\n\x07\x63ontent\x18\x03 \x01(\t\x12&\n\x07\x61\x64\x64ress\x18\x04 \x01(\x0b\x32\x15.syft.core.io.Address"\xf8\x01\n\x12UpdateSetupMessage\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12&\n\x07\x61\x64\x64ress\x18\x02 \x01(\x0b\x32\x15.syft.core.io.Address\x12\x13\n\x0b\x64omain_name\x18\x03 \x01(\t\x12\x13\n\x0b\x64\x65scription\x18\x04 \x01(\t\x12\x0b\n\x03\x64\x61\x61\x18\x05 \x01(\x08\x12\x0f\n\x07\x63ontact\x18\x06 \x01(\t\x12\x14\n\x0c\x64\x61\x61_document\x18\x08 \x01(\x0c\x12\x0c\n\x04tags\x18\t \x03(\t\x12\'\n\x08reply_to\x18\n \x01(\x0b\x32\x15.syft.core.io.Address"u\n\x13UpdateSetupResponse\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12\x0f\n\x07\x63ontent\x18\x03 \x01(\t\x12&\n\x07\x61\x64\x64ress\x18\x04 \x01(\x0b\x32\x15.syft.core.io.Addressb\x06proto3'
)

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(
    DESCRIPTOR, "proto.grid.messages.setup_messages_pb2", globals()
)
if _descriptor._USE_C_DESCRIPTORS == False:

    DESCRIPTOR._options = None
    _CREATEINITIALSETUPMESSAGE._serialized_start = 133
    _CREATEINITIALSETUPMESSAGE._serialized_end = 385
    _GETSETUPMESSAGE._serialized_start = 388
    _GETSETUPMESSAGE._serialized_end = 525
    _GETSETUPRESPONSE._serialized_start = 528
    _GETSETUPRESPONSE._serialized_end = 663
    _UPDATESETUPMESSAGE._serialized_start = 666
    _UPDATESETUPMESSAGE._serialized_end = 914
    _UPDATESETUPRESPONSE._serialized_start = 916
    _UPDATESETUPRESPONSE._serialized_end = 1033
# @@protoc_insertion_point(module_scope)



# future
from __future__ import annotations

# stdlib
import asyncio
import os
import time
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple
from typing import Union

# third party
import ascii_magic
from nacl.signing import SigningKey
from nacl.signing import VerifyKey
from pydantic import BaseSettings

# relative
from ...lib.python import String
from ...logger import critical
from ...logger import debug
from ...logger import info
from ...logger import traceback
from ..adp.ledger_store import RedisLedgerStore
from ..common.message import SignedImmediateSyftMessageWithReply
from ..common.message import SignedMessage
from ..common.message import SyftMessage
from ..common.uid import UID
from ..io.location import Location
from ..io.location import SpecificLocation
from .common.action.get_object_action import GetObjectAction
from .common.client import Client
from .common.node import Node
from .common.node_manager.association_request_manager import AssociationRequestManager
from .common.node_manager.dataset_manager import DatasetManager
from .common.node_manager.environment_manager import EnvironmentManager
from .common.node_manager.node_manager import NodeManager
from .common.node_manager.node_route_manager import NodeRouteManager
from .common.node_manager.redis_store import RedisStore
from .common.node_manager.request_manager import RequestManager
from .common.node_manager.role_manager import RoleManager
from .common.node_manager.user_manager import UserManager
from .common.node_service.association_request.association_request_service import (
    AssociationRequestService,
)
from .common.node_service.association_request.association_request_service import (
    AssociationRequestWithoutReplyService,
)
from .common.node_service.dataset_manager.dataset_manager_service import (
    DatasetManagerService,
)
from .common.node_service.get_remaining_budget.get_remaining_budget_service import (
    GetRemainingBudgetService,
)
from .common.node_service.node_setup.node_setup_messages import (
    CreateInitialSetUpMessage,
)
from .common.node_service.node_setup.node_setup_service import NodeSetupService
from .common.node_service.object_request.object_request_service import (
    ObjectRequestServiceWithoutReply,
)
from .common.node_service.object_request.object_request_service import RequestService
from .common.node_service.ping.ping_service import PingService
from .common.node_service.publish.publish_service import PublishScalarsService
from .common.node_service.request_answer.request_answer_messages import RequestStatus
from .common.node_service.request_answer.request_answer_service import (
    RequestAnswerService,
)
from .common.node_service.request_receiver.request_receiver_messages import (
    RequestMessage,
)
from .common.node_service.role_manager.role_manager_service import RoleManagerService
from .common.node_service.simple.simple_service import SimpleService
from .common.node_service.sleep.sleep_service import SleepService
from .common.node_service.user_auth.user_auth_service import UserLoginService
from .common.node_service.user_manager.user_manager_service import UserManagerService
from .common.node_service.vpn.vpn_service import VPNConnectService
from .common.node_service.vpn.vpn_service import VPNJoinService
from .common.node_service.vpn.vpn_service import VPNStatusService
from .common.node_table.utils import create_memory_db_engine
from .device import Device
from .device import DeviceClient
from .domain_client import DomainClient
from .domain_service import DomainServiceClass


class Domain(Node):
    domain: SpecificLocation
    root_key: Optional[VerifyKey]

    child_type = Device
    client_type = DomainClient
    child_type_client_type = DeviceClient

    def __init__(
        self,
        name: Optional[str],
        network: Optional[Location] = None,
        domain: SpecificLocation = SpecificLocation(),
        device: Optional[Location] = None,
        vm: Optional[Location] = None,
        signing_key: Optional[SigningKey] = None,
        verify_key: Optional[VerifyKey] = None,
        root_key: Optional[VerifyKey] = None,
        db_engine: Any = None,
        store_type: type = RedisStore,
        ledger_store_type: type = RedisLedgerStore,
        settings: Optional[BaseSettings] = None,
    ):

        if db_engine is None:
            db_engine, _ = create_memory_db_engine()

        super().__init__(
            name=name,
            network=network,
            domain=domain,
            device=device,
            vm=vm,
            signing_key=signing_key,
            verify_key=verify_key,
            db_engine=db_engine,
            store_type=store_type,
            settings=settings,
        )

        # share settings with the FastAPI application level
        self.settings = settings

        # specific location with name
        self.domain = SpecificLocation(name=self.name)
        self.root_key = root_key

        # Database Management Instances
        self.users = UserManager(db_engine)
        self.roles = RoleManager(db_engine)
        self.environments = EnvironmentManager(db_engine)
        self.association_requests = AssociationRequestManager(db_engine)
        self.data_requests = RequestManager(db_engine)
        self.datasets = DatasetManager(db_engine)
        self.node = NodeManager(db_engine)
        self.node_route = NodeRouteManager(db_engine)
        self.ledger_store = ledger_store_type(settings=settings)

        # self.immediate_services_without_reply.append(RequestReceiverService)
        # self.immediate_services_without_reply.append(AcceptOrDenyRequestService)
        # self.immediate_services_without_reply.append(UpdateRequestHandlerService)
        self.immediate_services_without_reply.append(PublishScalarsService)
        self.immediate_services_with_reply.append(RequestAnswerService)
        # self.immediate_services_with_reply.append(GetAllRequestHandlersService)

        # Grid Domain Services
        self.immediate_services_with_reply.append(AssociationRequestService)
        # self.immediate_services_with_reply.append(DomainInfrastructureService)
        self.immediate_services_with_reply.append(GetRemainingBudgetService)
        self.immediate_services_with_reply.append(SimpleService)
        self.immediate_services_with_reply.append(PingService)
        self.immediate_services_with_reply.append(VPNConnectService)
        self.immediate_services_with_reply.append(VPNJoinService)
        self.immediate_services_with_reply.append(VPNStatusService)
        self.immediate_services_with_reply.append(NodeSetupService)
        self.immediate_services_with_reply.append(RoleManagerService)
        self.immediate_services_with_reply.append(UserManagerService)
        self.immediate_services_with_reply.append(DatasetManagerService)
        self.immediate_services_with_reply.append(RequestService)
        self.immediate_services_with_reply.append(UserLoginService)

        self.immediate_services_without_reply.append(ObjectRequestServiceWithoutReply)
        self.immediate_services_without_reply.append(
            AssociationRequestWithoutReplyService
        )

        # TODO: New Service registration process
        self.immediate_services_with_reply.append(DomainServiceClass)

        # TODO: @Madhava change to a map of accountants that are created on first
        # use of the DS key

        if getattr(self.settings, "TEST_MODE", False):
            print("Loading TEST_MODE services")
            # only add in test mode
            self.immediate_services_with_reply.append(SleepService)

        self.requests: List[RequestMessage] = list()
        # available_device_types = set()
        # TODO: add available compute types

        # default_device = None
        # TODO: add default compute type

        self._register_services()
        self.request_handlers: List[Dict[Union[str, String], Any]] = []
        self.handled_requests: Dict[Any, float] = {}

        self.post_init()

        # run the handlers in an asyncio future
        asyncio.ensure_future(self.run_handlers())

    def post_init(self) -> None:
        super().post_init()
        self.set_node_uid()
        if not hasattr(self, "signing_key"):
            Node.set_keys(node=self)

    def initial_setup(  # nosec
        self,
        signing_key: SigningKey,
        first_superuser_name: str = "Jane Doe",
        first_superuser_email: str = "info@openmined.org",
        first_superuser_password: str = "changethis",
        first_superuser_budget: float = 5.55,
        domain_name: str = "BigHospital",
    ) -> Domain:
        Node.set_keys(node=self, signing_key=signing_key)

        # Build Syft Message
        msg: SignedImmediateSyftMessageWithReply = CreateInitialSetUpMessage(
            address=self.address,
            name=first_superuser_name,
            email=first_superuser_email,
            password=first_superuser_password,
            domain_name=domain_name,
            budget=first_superuser_budget,
            reply_to=self.address,
            signing_key=signing_key,
        ).sign(signing_key=self.signing_key)

        # Process syft message
        _ = self.recv_immediate_msg_with_reply(msg=msg).message

        return self

    def loud_print(self) -> None:
        try:
            install_path = os.path.abspath(
                os.path.join(os.path.realpath(__file__), "../../../../img/")
            )
            ascii_magic.to_terminal(
                ascii_magic.from_image_file(
                    img_path=install_path + "/pygrid.png", columns=83
                )
            )

            print(
                r"""
                                                         __
                                                        |  \  _   _   _  .  _
                                                        |__/ (_) ||| (_| | | )
    """
            )
        except Exception:
            print("DOMAIN (broken file path)")

    @property
    def icon(self) -> str:
        return "🏰"

    @property
    def id(self) -> UID:
        return self.domain.id

    def message_is_for_me(self, msg: Union[SyftMessage, SignedMessage]) -> bool:

        # this needs to be defensive by checking domain_id NOT domain.id or it breaks
        try:
            return msg.address.domain_id == self.id and msg.address.device is None
        except Exception as excp3:
            critical(
                f"Error checking if {msg.pprint} is for me on {self.pprint}. {excp3}"
            )
            return False

    def set_request_status(
        self, message_request_id: UID, status: RequestStatus, client: Client
    ) -> bool:
        for req in self.requests:
            if req.request_id == message_request_id:
                req.owner_client_if_available = client
                if status == RequestStatus.Accepted:
                    req.accept()
                    return True
                elif status == RequestStatus.Rejected:
                    req.deny()
                    return True

        return False

    def get_request_status(self, message_request_id: UID) -> RequestStatus:
        # is it still pending
        for req in self.requests:
            if req.request_id == message_request_id:
                return RequestStatus.Pending

        # check if it was accepted
        # TODO remove brute search of all store objects
        # Currently theres no way to find which object to check the permissions
        # to find the stored request_id
        for obj_id in self.store.keys():
            for _, request_id in self.store.get(
                obj_id, proxy_only=True
            ).read_permissions.items():
                if request_id == message_request_id:
                    return RequestStatus.Accepted

            for _, request_id in self.store.get(
                obj_id, proxy_only=True
            ).search_permissions.items():
                if request_id == message_request_id:
                    return RequestStatus.Accepted

        # must have been rejected
        return RequestStatus.Rejected

    def _get_object(self, request: RequestMessage) -> Optional[Any]:
        try:
            obj_msg = GetObjectAction(
                id_at_location=request.object_id,
                address=request.owner_address,
                reply_to=self.address,
                delete_obj=False,
            )

            service = self.immediate_msg_with_reply_router[type(obj_msg)]
            response = service.process(
                node=self, msg=obj_msg, verify_key=self.root_verify_key
            )
            if response:
                obj = getattr(response, "obj", None)
                if obj is not None:
                    return obj
        except Exception as excp1:
            critical(f"Exception getting object for {request}. {excp1}")
        return None

    def _count_elements(self, obj: object) -> Tuple[bool, int]:
        allowed = False
        elements = 0

        nelement = getattr(obj, "nelement", None)
        if nelement is not None:
            elements = max(elements, int(nelement()))
            allowed = True

        return (allowed, elements)

    def _accept(self, request: RequestMessage) -> None:
        debug(f"Calling accept on request: {request.id}")
        request.destination_node_if_available = self
        request.accept()

    def _deny(self, request: RequestMessage) -> None:
        debug(f"Calling deny on request: {request.id}")
        request.destination_node_if_available = self
        request.deny()

    def _try_deduct_quota(
        self, handler: Dict[Union[str, String], Any], obj: Any
    ) -> bool:
        action = handler.get("action", None)
        if action == "accept":
            allowed, element_count = self._count_elements(obj=obj)
            if allowed:
                result = handler["element_quota"] - element_count
                if result >= 0:
                    # the request will be accepted so lets decrement the quota
                    handler["element_quota"] = max(0, result)
                    return True

        return False

    def check_handler(
        self, handler: Dict[Union[str, String], Any], request: RequestMessage
    ) -> bool:
        debug(f"HANDLER Check handler {handler} against {request.request_id}")

        tags = handler.get("tags", [])

        action = handler.get("action", None)
        print_local = handler.get("print_local", None)
        log_local = handler.get("log_local", None)
        element_quota = handler.get("element_quota", None)

        # We match a handler and a request when they have a same set of tags,
        # or if handler["tags"]=[], it matches with any request.
        if len(tags) > 0 and not set(request.object_tags) == set(tags):
            debug(f"HANDLER Ignoring request handler {handler} against {request}")
            return False

        # if we have any of these three rules we will need to get the object to
        # print it, log it, or check its quota
        obj = None
        if print_local or log_local or element_quota:
            obj = self._get_object(request=request)
            debug(f"> HANDLER Got object {obj} for checking")

        # we only want to accept or deny once
        handled = False

        # check quota and reject first
        if element_quota is not None:
            if not self._try_deduct_quota(handler=handler, obj=obj):
                debug(
                    f"> HANDLER Rejecting {request} element_quota={handler['element_quota']}"
                )
                self._deny(request=request)
                handled = True

        # if not rejected based on quota keep checking
        if not handled:
            if action == "accept":
                debug(f"Check accept {handler} against {request}")
                self._accept(request=request)
                handled = True
            elif action == "deny":
                self._deny(request=request)
                handled = True

        # print or log rules can execute multiple times so no complex logic here
        if print_local or log_local:
            log = f"> HANDLER Request {request.request_id}:"
            if len(request.request_description) > 0:
                log += f" {request.request_description}"
            log += f"\nValue: {obj}"

            # if these are enabled output them
            if print_local:
                critical(log)

            if log_local:
                info(log)

        # block the loop from handling this again, until the cleanup removes it
        # after a period of timeout
        if handled:
            self.handled_requests[request.id] = time.time()
        return handled

    def clean_up_handlers(self) -> None:
        # this makes sure handlers with timeout expire
        now = time.time()
        alive_handlers = []
        if len(self.request_handlers) > 0:
            for handler in self.request_handlers:
                timeout_secs = handler.get("timeout_secs", -1)
                if timeout_secs != -1:
                    created_time = handler.get("created_time", 0)
                    if now - created_time > timeout_secs:
                        continue
                alive_handlers.append(handler)
        self.request_handlers = alive_handlers

    def clear(self, user_role: int) -> bool:
        # Cleanup database tables
        if user_role == self.roles.owner_role.id:
            self.store.clear()
            self.data_requests.clear()
            self.users.clear()
            self.environments.clear()
            self.association_requests.clear()
            self.datasets.clear()
            self.initial_setup(signing_key=self.signing_key)
            return True

        return False

    def clean_up_requests(self) -> None:
        # this allows a request to be re-handled if the handler somehow failed
        now = time.time()
        processing_wait_secs = 5
        reqs_to_remove = []
        for req in self.handled_requests.keys():
            handle_time = self.handled_requests[req]
            if now - handle_time > processing_wait_secs:
                reqs_to_remove.append(req)

        for req in reqs_to_remove:
            self.handled_requests.__delitem__(req)

        alive_requests: List[RequestMessage] = []
        for request in self.requests:
            if request.timeout_secs is not None and request.timeout_secs > -1:
                if request.arrival_time is None:
                    critical(f"HANDLER Request has no arrival time. {request.id}")
                    request.set_arrival_time(arrival_time=time.time())
                arrival_time = getattr(request, "arrival_time", float(now))
                if now - arrival_time > request.timeout_secs:
                    # this request has expired
                    continue
            alive_requests.append(request)

        self.requests = alive_requests

    async def run_handlers(self) -> None:
        while True:
            await asyncio.sleep(0.01)
            try:
                self.clean_up_handlers()
                self.clean_up_requests()
                if len(self.request_handlers) > 0:
                    for request in self.requests:
                        # check if we have previously already handled this in an earlier iter
                        if request.id not in self.handled_requests:
                            for handler in self.request_handlers:
                                handled = self.check_handler(
                                    handler=handler, request=request
                                )
                                if handled:
                                    # we handled the request so we can exit the loop
                                    break
            except Exception as excp2:
                traceback(excp2)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: proto/grid/messages/tensor_messages.proto
"""Generated protocol buffer code."""
# third party
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


# syft absolute
from syft.proto.core.common import (
    common_object_pb2 as proto_dot_core_dot_common_dot_common__object__pb2,
)
from syft.proto.core.io import address_pb2 as proto_dot_core_dot_io_dot_address__pb2

DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n)proto/grid/messages/tensor_messages.proto\x12\x12syft.grid.messages\x1a%proto/core/common/common_object.proto\x1a\x1bproto/core/io/address.proto"\x9e\x01\n\x13\x43reateTensorMessage\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12&\n\x07\x61\x64\x64ress\x18\x02 \x01(\x0b\x32\x15.syft.core.io.Address\x12\x0f\n\x07\x63ontent\x18\x03 \x01(\t\x12\'\n\x08reply_to\x18\x04 \x01(\x0b\x32\x15.syft.core.io.Address"\x8b\x01\n\x14\x43reateTensorResponse\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12\x13\n\x0bstatus_code\x18\x02 \x01(\x05\x12\x0f\n\x07\x63ontent\x18\x03 \x01(\t\x12&\n\x07\x61\x64\x64ress\x18\x04 \x01(\x0b\x32\x15.syft.core.io.Address"\x9b\x01\n\x10GetTensorMessage\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12&\n\x07\x61\x64\x64ress\x18\x02 \x01(\x0b\x32\x15.syft.core.io.Address\x12\x0f\n\x07\x63ontent\x18\x03 \x01(\t\x12\'\n\x08reply_to\x18\x04 \x01(\x0b\x32\x15.syft.core.io.Address"\x88\x01\n\x11GetTensorResponse\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12\x13\n\x0bstatus_code\x18\x02 \x01(\x05\x12\x0f\n\x07\x63ontent\x18\x03 \x01(\t\x12&\n\x07\x61\x64\x64ress\x18\x04 \x01(\x0b\x32\x15.syft.core.io.Address"\x9c\x01\n\x11GetTensorsMessage\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12&\n\x07\x61\x64\x64ress\x18\x02 \x01(\x0b\x32\x15.syft.core.io.Address\x12\x0f\n\x07\x63ontent\x18\x03 \x01(\t\x12\'\n\x08reply_to\x18\x04 \x01(\x0b\x32\x15.syft.core.io.Address"\x89\x01\n\x12GetTensorsResponse\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12\x13\n\x0bstatus_code\x18\x02 \x01(\x05\x12\x0f\n\x07\x63ontent\x18\x03 \x01(\t\x12&\n\x07\x61\x64\x64ress\x18\x04 \x01(\x0b\x32\x15.syft.core.io.Address"\x9e\x01\n\x13\x44\x65leteTensorMessage\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12&\n\x07\x61\x64\x64ress\x18\x02 \x01(\x0b\x32\x15.syft.core.io.Address\x12\x0f\n\x07\x63ontent\x18\x03 \x01(\t\x12\'\n\x08reply_to\x18\x04 \x01(\x0b\x32\x15.syft.core.io.Address"\x8b\x01\n\x14\x44\x65leteTensorResponse\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12\x13\n\x0bstatus_code\x18\x02 \x01(\x05\x12\x0f\n\x07\x63ontent\x18\x03 \x01(\t\x12&\n\x07\x61\x64\x64ress\x18\x04 \x01(\x0b\x32\x15.syft.core.io.Address"\x9e\x01\n\x13UpdateTensorMessage\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12&\n\x07\x61\x64\x64ress\x18\x02 \x01(\x0b\x32\x15.syft.core.io.Address\x12\x0f\n\x07\x63ontent\x18\x03 \x01(\t\x12\'\n\x08reply_to\x18\x04 \x01(\x0b\x32\x15.syft.core.io.Address"\x8b\x01\n\x14UpdateTensorResponse\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12\x13\n\x0bstatus_code\x18\x02 \x01(\x05\x12\x0f\n\x07\x63ontent\x18\x03 \x01(\t\x12&\n\x07\x61\x64\x64ress\x18\x04 \x01(\x0b\x32\x15.syft.core.io.Addressb\x06proto3'
)

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(
    DESCRIPTOR, "proto.grid.messages.tensor_messages_pb2", globals()
)
if _descriptor._USE_C_DESCRIPTORS == False:

    DESCRIPTOR._options = None
    _CREATETENSORMESSAGE._serialized_start = 134
    _CREATETENSORMESSAGE._serialized_end = 292
    _CREATETENSORRESPONSE._serialized_start = 295
    _CREATETENSORRESPONSE._serialized_end = 434
    _GETTENSORMESSAGE._serialized_start = 437
    _GETTENSORMESSAGE._serialized_end = 592
    _GETTENSORRESPONSE._serialized_start = 595
    _GETTENSORRESPONSE._serialized_end = 731
    _GETTENSORSMESSAGE._serialized_start = 734
    _GETTENSORSMESSAGE._serialized_end = 890
    _GETTENSORSRESPONSE._serialized_start = 893
    _GETTENSORSRESPONSE._serialized_end = 1030
    _DELETETENSORMESSAGE._serialized_start = 1033
    _DELETETENSORMESSAGE._serialized_end = 1191
    _DELETETENSORRESPONSE._serialized_start = 1194
    _DELETETENSORRESPONSE._serialized_end = 1333
    _UPDATETENSORMESSAGE._serialized_start = 1336
    _UPDATETENSORMESSAGE._serialized_end = 1494
    _UPDATETENSORRESPONSE._serialized_start = 1497
    _UPDATETENSORRESPONSE._serialized_end = 1636
# @@protoc_insertion_point(module_scope)



# stdlib
from typing import Optional

# third party
from nacl.encoding import HexEncoder
from nacl.signing import VerifyKey
from sqlalchemy.engine import Engine

# relative
from ..node_service.node_credential.node_credentials import NodeCredentials
from ..node_table.node import Node
from .database_manager import DatabaseManager


class NodeManager(DatabaseManager):
    schema = Node

    def __init__(self, database: Engine) -> None:
        super().__init__(schema=NodeManager.schema, db=database)

    def create_or_get_node(self, node_uid: str, node_name: str) -> int:
        # node_uid is a UID as a string with no_dash
        node = self.first(node_uid=node_uid)
        if node:
            self.modify(query={"node_uid": node_uid}, values={"node_name": node_name})
        else:
            self.register(**{"node_uid": node_uid, "node_name": node_name})
            node = self.first(node_uid=node_uid)
        return node.id

    def add_or_update_node_credentials(self, credentials: NodeCredentials) -> None:
        node = self.first(node_uid=credentials.node_uid)
        if node:
            if node.verify_key is not None:
                credentials.validate(key=node.verify_key)
            self.modify(
                query={"node_uid": credentials.node_uid}, values={**credentials}
            )
        else:
            self.register(**credentials)

    def validate_id_and_key(
        self, node_uid: str, verify_key: VerifyKey
    ) -> Optional[Node]:
        return self.first(
            node_uid=node_uid,
            verify_key=verify_key.encode(encoder=HexEncoder).decode("utf-8"),
        )

    def get_node_for(self, verify_key: VerifyKey) -> Optional[Node]:
        return self.first(
            verify_key=verify_key.encode(encoder=HexEncoder).decode("utf-8")
        )



# future
from __future__ import annotations

# stdlib
from typing import Any

# third party
from google.protobuf.reflection import GeneratedProtocolMessageType

# relative
from ...core.common.serde.serializable import serializable
from ...proto.lib.python.bytes_pb2 import Bytes as Bytes_PB
from .primitive_interface import PyPrimitive


@serializable()
class Bytes(bytes, PyPrimitive):
    def __new__(cls, value: Any = None) -> Bytes:
        if value is None:
            value = b""

        return bytes.__new__(cls, value)

    def __init__(self, value: Any = None):
        if value is None:
            value = 0

        bytes.__init__(value)

    def upcast(self) -> bytes:
        return bytes(self)

    def _object2proto(self) -> Bytes_PB:
        bytes_pb = Bytes_PB()
        bytes_pb.data = self
        return bytes_pb

    @staticmethod
    def _proto2object(proto: Bytes_PB) -> Bytes:
        return Bytes(value=proto.data)

    @staticmethod
    def get_protobuf_schema() -> GeneratedProtocolMessageType:
        return Bytes_PB



# stdlib
from typing import List

# third party
from nacl.signing import VerifyKey
from sqlalchemy.engine import Engine

# relative
from .....grid import GridURL
from ..node_service.node_route.route_update import RouteUpdate
from ..node_table.node import Node as NodeRow
from ..node_table.node_route import NodeRoute
from .database_manager import DatabaseManager


class NodeRouteManager(DatabaseManager):
    schema = NodeRoute

    def __init__(self, database: Engine) -> None:
        super().__init__(schema=NodeRouteManager.schema, db=database)

    def update_route_for_node(
        self, node_id: int, host_or_ip: str, is_vpn: bool = False
    ) -> NodeRoute:
        # node_id is a database int id
        # host_or_ip can have a port as well

        node_route = self.first(host_or_ip=host_or_ip)
        if node_route:
            self.modify(
                query={"host_or_ip": host_or_ip},
                values={"is_vpn": is_vpn, "node_id": node_id},
            )
        else:
            self.register(
                **{"node_id": node_id, "host_or_ip": host_or_ip, "is_vpn": is_vpn}
            )
            node_route = self.first(host_or_ip=host_or_ip)
        return node_route

    def validate_route_update(
        self, node_row: NodeRow, route_update: RouteUpdate, verify_key: VerifyKey
    ) -> bool:
        if not route_update.source_node_url:
            raise Exception("source_node_url is missing")
        source_url = GridURL.from_url(route_update.source_node_url)

        node_route = self.first(host_or_ip=source_url.host_or_ip, port=source_url.port)
        if node_route is None:
            # not created yet so valid
            return True
        elif node_route and node_route.node_id == node_row.id:
            # matches so valid
            return True
        # someone else already has this host and port so return False
        return False

    def update_route(
        self, node_row: NodeRow, route_update: RouteUpdate, is_vpn: bool = False
    ) -> NodeRoute:
        if not route_update.source_node_url:
            raise Exception("source_node_url is missing")
        source_url = GridURL.from_url(route_update.source_node_url)

        node_route = self.first(host_or_ip=source_url.host_or_ip, node_id=node_row.id)
        if node_route:
            self.modify(
                query={"host_or_ip": source_url.host_or_ip, "node_id": node_row.id},
                values={
                    "is_vpn": is_vpn,
                    "private": route_update.private,
                    "protocol": source_url.protocol,
                    "port": source_url.port,
                },
            )
        else:
            self.register(
                **{
                    "is_vpn": is_vpn,
                    "node_id": node_row.id,
                    "private": route_update.private,
                    "protocol": source_url.protocol,
                    "port": source_url.port,
                    "host_or_ip": source_url.host_or_ip,
                }
            )
        new_node_route_row = self.first(
            host_or_ip=source_url.host_or_ip, node_id=node_row.id
        )
        if new_node_route_row:
            return new_node_route_row
        raise Exception("Failed to update node_route")

    def get_routes(self, node_row: NodeRow) -> List[NodeRoute]:
        return self.query(node_id=node_row.id)



"""The GC class that would handle what happens when a pointer gets deleted."""
# stdlib
from typing import Any
from typing import Dict
from typing import List

# third party
from typing_extensions import final

# relative
from ..pointer import Pointer
from .gc_strategy import GCStrategy


@final
class GarbageCollection:
    __slots__ = ["_gc_strategy"]

    _gc_strategy: GCStrategy

    def __init__(
        self, gc_strategy_name: str, *args: List[Any], **kwargs: Dict[Any, Any]
    ) -> None:
        """Initialize a strategy given the name of it.

        Args:
            gc_strategy_name (str): name of the registered GC strategy
            *args (List[Any]): args to be passed to the GCStrategy init method
            **kwargs (Dict[Any, Any]): kwargs to be passed to the GCStrategy
                            init method
        Return:
            None
        """
        if gc_strategy_name not in GCStrategy.REGISTERED_GC_STRATEGIES:
            raise ValueError(f"{gc_strategy_name} not registered!")

        gc_strategy_init = GCStrategy.REGISTERED_GC_STRATEGIES[gc_strategy_name]
        self._gc_strategy = gc_strategy_init(*args, **kwargs)

    def apply(self, pointer: Pointer) -> None:
        """Apply the GCStrategy garbage collection logic on the pointer.

        Args:
            pointer (Pointer): the pointer that should be garbage collected

        Return:
            None
        """
        self._gc_strategy.reap(pointer)

    @property
    def gc_strategy(self) -> GCStrategy:
        """Get the GCStrategy.

        Return:
            The used GCStrategy
        """
        return self._gc_strategy

    @gc_strategy.setter
    def gc_strategy(self, strategy: GCStrategy) -> None:
        """Set the GCStrategy.

        Args:
            strategy (GCStrategy): The GCStrategy to be set

        Return:
            None
        """
        self._gc_strategy = strategy



# future
from __future__ import annotations

# stdlib
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import cast

# third party
from nacl.signing import VerifyKey
from typing_extensions import final

# relative
from .....common.serde.serializable import serializable
from ....abstract.node_service_interface import NodeServiceInterface
from ...node_table.node import Node as NodeRow
from ...node_table.node_route import NodeRoute as NodeRouteRow
from ..generic_payload.messages import GenericPayloadMessage
from ..generic_payload.messages import GenericPayloadMessageWithReply
from ..generic_payload.messages import GenericPayloadReplyMessage


@serializable(recursive_serde=True)
@final
class PeerDiscoveryMessage(GenericPayloadMessage):
    ...


@serializable(recursive_serde=True)
@final
class PeerDiscoveryReplyMessage(GenericPayloadReplyMessage):
    ...


def node_id_to_peer_route_metadata(
    node: NodeServiceInterface, node_row: NodeRow
) -> List[Dict[str, Any]]:
    routes: List[NodeRouteRow] = node.node_route.query(node_id=node_row.id)
    routes_meta = []
    for route in routes:
        peer_route = {}
        peer_route["id"] = node_row.node_uid
        peer_route["name"] = node_row.node_name
        peer_route["host_or_ip"] = route.host_or_ip
        peer_route["is_vpn"] = route.is_vpn
        peer_route["private"] = route.private
        peer_route["protocol"] = route.protocol
        peer_route["port"] = route.port
        routes_meta.append(peer_route)
    return routes_meta


@serializable(recursive_serde=True)
@final
class PeerDiscoveryMessageWithReply(GenericPayloadMessageWithReply):
    message_type = PeerDiscoveryMessage
    message_reply_type = PeerDiscoveryReplyMessage

    def run(
        self, node: NodeServiceInterface, verify_key: Optional[VerifyKey] = None
    ) -> Dict[str, Any]:
        try:
            peer_routes = []
            for peer in node.node.all():
                peer_routes += node_id_to_peer_route_metadata(node=node, node_row=peer)
            return {"status": "ok", "data": peer_routes}
        except Exception as e:
            print(f"Failed to run {type(self)}", self.kwargs, e)
            return {"status": "error"}


@serializable(recursive_serde=True)
@final
class GetPeerInfoMessageWithReply(GenericPayloadMessageWithReply):
    message_type = PeerDiscoveryMessage
    message_reply_type = PeerDiscoveryReplyMessage

    def run(
        self, node: NodeServiceInterface, verify_key: Optional[VerifyKey] = None
    ) -> Dict[str, Any]:
        try:
            # TODO: handle multiple routes and possibly prefer vpn connections
            peer = node.node.first(node_uid=str(self.kwargs["uid"]))
            peer_route = {}
            if peer:
                peer = cast(NodeRow, peer)
                route = node.node_route.first(node_id=peer.id)
                if route:
                    route = cast(NodeRouteRow, route)
                    peer_route["id"] = peer.node_uid
                    peer_route["name"] = peer.node_name
                    peer_route["host_or_ip"] = route.host_or_ip
                    peer_route["is_vpn"] = route.is_vpn
            return {"status": "ok", "data": peer_route}
        except Exception as e:
            print(f"Failed to run {type(self)}", self.kwargs, e)
            return {"status": "error"}



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: proto/lib/torch/size.proto
"""Generated protocol buffer code."""
# third party
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


# syft absolute
from syft.proto.core.common import (
    common_object_pb2 as proto_dot_core_dot_common_dot_common__object__pb2,
)

DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\x1aproto/lib/torch/size.proto\x12\x0esyft.lib.torch\x1a%proto/core/common/common_object.proto"7\n\x04Size\x12\x0c\n\x04\x64\x61ta\x18\x01 \x03(\x04\x12!\n\x02id\x18\x02 \x01(\x0b\x32\x15.syft.core.common.UIDb\x06proto3'
)

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(
    DESCRIPTOR, "proto.lib.torch.size_pb2", globals()
)
if _descriptor._USE_C_DESCRIPTORS == False:

    DESCRIPTOR._options = None
    _SIZE._serialized_start = 85
    _SIZE._serialized_end = 140
# @@protoc_insertion_point(module_scope)



# third party
from sqlalchemy import Column
from sqlalchemy import ForeignKey
from sqlalchemy import Integer
from sqlalchemy import String

# relative
from . import Base


class Ledger(Base):
    __tablename__ = "ledger"

    id = Column(Integer(), primary_key=True, autoincrement=True)
    entity_name = Column(String(256), ForeignKey("entity.name", ondelete="CASCADE"))
    mechanism_id = Column(Integer(), ForeignKey("mechanism.id", ondelete="CASCADE"))



# relative
from .abstract_node_msg_registry import AbstractNodeMessageRegistry


class NetworkMessageRegistry(AbstractNodeMessageRegistry):
    """A class for registering messages that can be used by the network service."""



"""Function used to calculate exp of a given tensor."""
# future
from __future__ import annotations

# stdlib
from typing import TYPE_CHECKING
from typing import Union

if TYPE_CHECKING:
    # relative
    from ...tensor import Tensor
    from ...tensor.smpc.mpc_tensor import MPCTensor


def exp(
    value: Union[Tensor, MPCTensor, int, float], iterations: int = 8
) -> Union[MPCTensor, float, Tensor]:
    """Approximates the exponential function using a limit approximation.

    exp(x) = lim_{n -> infty} (1 + x / n) ^ n
    Here we compute exp by choosing n = 2 ** d for some large d equal to
    iterations. We then compute (1 + x / n) once and square `d` times.

    Args:
        value: tensor whose exp is to be calculated
        iterations (int): number of iterations for limit approximation
    Ref: https://github.com/LaRiffle/approximate-models

    Returns:
        MPCTensor: the calculated exponential of the given tensor
    """
    result = (value / 2**iterations) + 1
    for _ in range(iterations):
        result = result * result
    return result



# relative
from .activations import leaky_ReLU
from .layers.convolution import Convolution
from .layers.linear import Linear
from .layers.normalization import BatchNorm
from .layers.pooling import AvgPool
from .layers.pooling import MaxPool
from .layers.reshaping import Flatten
from .loss import BinaryCrossEntropy
from .model import Model

__all__ = [
    "Linear",
    "BinaryCrossEntropy",
    "Convolution",
    "BatchNorm",
    "AvgPool",
    "MaxPool",
    "Model",
    "leaky_ReLU",
    "Flatten",
]



class ADPTensor(object):
    """This tensor is just here to be an abstract class for now so that we can do
    isinstance() checks on DP tensors without having to worry about each one
    individually. We might add generic functionality to this in the future though."""



# third party
from google.protobuf.message import Message

# relative
from ....logger import traceback_and_raise
from ....proto.util.data_message_pb2 import DataMessage
from ....util import get_fully_qualified_name
from ....util import validate_type
from .deserialize import PROTOBUF_START_MAGIC_HEADER
from .types import Deserializeable


def create_protobuf_magic_header() -> str:
    return f"{PROTOBUF_START_MAGIC_HEADER}"


def _serialize(
    obj: object,
    to_proto: bool = True,
    to_bytes: bool = False,
) -> Deserializeable:
    """Serialize the object according to the parameters.

    This method can be called directly on the syft module::

        import syft as sy
        serialized_obj = sy.serialize(obj=my_object_to_serialize)

    This is the primary serialization method, which processes the above
    flags in a particular order. In general, it is not expected that people
    will set multiple to_<type> flags to True at the same time. We don't
    currently have logic which prevents this, because this may affect
    runtime performance, but if several flags are True, then we will simply
    take return the type of latest supported flag from the following list:

        - proto
        - binary

    TODO: we could also add "dict" to this list but it's not clear if it would be used.

    :param to_proto: set this flag to TRUE if you want to return a protobuf object
    :type to_proto: bool
    :param to_bytes: set this flag to TRUE if you want to return a binary object
    :type to_bytes: bool
    :return: a serialized form of the object on which serialize() is called.
    :rtype: Union[str, bytes, Message]
    """

    # relative
    from ....lib.python.primitive_factory import isprimitive

    # we have an unboxed primitive type so we need to mirror that on deserialize
    if isprimitive(obj):
        # relative
        from ....lib.python.primitive_factory import PrimitiveFactory

        obj = PrimitiveFactory.generate_primitive(value=obj, temporary_box=True)
        if hasattr(obj, "temporary_box"):
            # TODO: can remove this once all of PrimitiveFactory.generate_primitive
            # supports temporary_box and is tested
            obj.temporary_box = True  # type: ignore

    if hasattr(obj, "_sy_serializable_wrapper_type"):
        is_serializable = obj._sy_serializable_wrapper_type(value=obj)  # type: ignore
    else:
        is_serializable = obj

    # traceback_and_raise(
    #     Exception(
    #         f"Object {type(obj)} is not serializable and has no _sy_serializable_wrapper_type"
    #     )
    # )

    # capnp_bytes=True
    if hasattr(is_serializable, "_object2bytes"):
        # capnp proto
        return validate_type(is_serializable._object2bytes(), bytes)

    if to_bytes:
        # debug(f"Serializing {type(is_serializable)}")
        # indent=None means no white space or \n in the serialized version
        # this is compatible with json.dumps(x, indent=None)
        serialized_data = is_serializable._object2proto().SerializeToString()
        obj_type = get_fully_qualified_name(obj=is_serializable)
        blob: Message = DataMessage(
            magic_header=create_protobuf_magic_header(),
            obj_type=obj_type,
            content=serialized_data,
        )
        return validate_type(blob.SerializeToString(), bytes)
    elif to_proto:
        return validate_type(is_serializable._object2proto(), Message)
    else:
        traceback_and_raise(
            Exception(
                """You must specify at least one deserialization format using
                        one of the arguments of the serialize() method such as:
                        to_proto, to_bytes."""
            )
        )



# relative
from .abstract_node_msg_registry import AbstractNodeMessageRegistry


class DomainMessageRegistry(AbstractNodeMessageRegistry):
    """A class for registering messages that can be used by the domain service."""



# future
from __future__ import annotations

# stdlib
import secrets
from typing import Callable
from typing import List
from typing import TYPE_CHECKING
from typing import Tuple
from typing import Union

# third party
import jax
from jax import numpy as jnp
import numpy as np
from tqdm import tqdm

# relative
from ..tensor.fixed_precision_tensor import FixedPrecisionTensor
from ..tensor.lazy_repeat_array import lazyrepeatarray
from ..tensor.passthrough import PassthroughTensor  # type: ignore
from .data_subject_ledger import DataSubjectLedger
from .data_subject_ledger import RDPParams
from .data_subject_list import DataSubjectList

if TYPE_CHECKING:
    # relative
    from ..tensor.autodp.gamma_tensor import GammaTensor

# def calculate_bounds_for_mechanism(
#     value_array: np.ndarray, min_val_array: np.ndarray, max_val_array: np.ndarray
# ) -> Tuple[np.ndarray, np.ndarray]:
#     """Calculates the squared L2 norm values needed to create a Mechanism, and calculate
#     privacy budget + spend. If you calculate the privacy budget spend with the worst
#     case bound, you can show this number to the DS. If you calculate it with the
#     regular value (the value computed below when public_only = False, you cannot show
#     the privacy budget to the DS because this violates privacy."""

#     # TODO: Double check whether the iDPGaussianMechanism class squares its
#     # squared_l2_norm values!!

#     # min_val_array = min_val_array.astype(np.int64)
#     # max_val_array = max_val_array.astype(np.int64)

#     # using np.ones_like dtype=value_array.dtype because without it the output was
#     # of type "O" python object causing issues when doing operations against JAX
#     worst_case_l2_norm = np.sqrt(
#         np.sum(np.square(max_val_array - min_val_array))
#     ) * np.ones_like(
#         value_array
#     )  # dtype=value_array.dtype)

#     l2_norm = np.sqrt(np.sum(np.square(value_array))) * np.ones_like(value_array)
#     # dtype=value_array.dtype
#     #
#     # print(l2_norm.shape, worst_case_l2_norm.shape)
#     # print(l2_norm.shape)
#     return l2_norm, worst_case_l2_norm


@jax.jit
def calculate_bounds_for_mechanism(
    value_array: jnp.ndarray,
    min_val_array: jnp.ndarray,
    max_val_array: jnp.ndarray,
    sigma: float,
) -> Tuple[jnp.ndarray, jnp.ndarray, jnp.array, jnp.array]:
    ones_like = jnp.ones_like(value_array)
    one_dim = jnp.reshape(ones_like, -1)

    worst_case_l2_norm = (
        jnp.sqrt(jnp.sum(jnp.square(max_val_array - min_val_array))) * one_dim
    )

    l2_norm = jnp.sqrt(jnp.sum(jnp.square(value_array))) * one_dim
    return l2_norm, worst_case_l2_norm, one_dim * sigma, one_dim


def vectorized_publish(
    min_vals: np.ndarray,
    max_vals: np.ndarray,
    state_tree: dict[int, "GammaTensor"],
    data_subjects: DataSubjectList,
    ledger: DataSubjectLedger,
    get_budget_for_user: Callable,
    deduct_epsilon_for_user: Callable,
    is_linear: bool = True,
    sigma: float = 1.5,
    output_func: Callable = lambda x: x,
) -> Union[np.ndarray, jax.numpy.DeviceArray]:
    # relative
    from ..tensor.autodp.gamma_tensor import GammaTensor

    # print(f"Starting vectorized publish: {type(ledger)}")
    # print("Starting RDP Params Calculation")
    # TODO: Vectorize this to return a larger GammaTensor instead of a list of Tensors
    input_tensors: List[GammaTensor] = GammaTensor.get_input_tensors(state_tree)

    filtered_inputs = []

    # This will reveal the # of input tensors to the user- remove this before merging to dev
    for input_tensor in tqdm(input_tensors):
        # TODO: Double check with Andrew if this is correct- if we use the individual min/max values

        # t1 = time()
        # Calculate everything needed for RDP

        if isinstance(input_tensor.child, FixedPrecisionTensor):
            value = input_tensor.child.decode()
        else:
            value = input_tensor.child

        while isinstance(value, PassthroughTensor):
            value = value.child

        if isinstance(input_tensor.min_vals, lazyrepeatarray):
            min_val_array = input_tensor.min_vals.to_numpy()
        else:
            min_val_array = input_tensor.min_vals

        if isinstance(input_tensor.max_vals, lazyrepeatarray):
            max_val_array = input_tensor.max_vals.to_numpy()
        else:
            max_val_array = input_tensor.max_vals

        l2_norms, l2_norm_bounds, sigmas, coeffs = calculate_bounds_for_mechanism(
            value_array=value,
            min_val_array=min_val_array,
            max_val_array=max_val_array,
            sigma=sigma,
        )

        if is_linear:
            lipschitz_bounds = coeffs.copy()
        else:
            lipschitz_bounds = input_tensor.lipschitz_bound
            # raise Exception("gamma_tensor.lipschitz_bound property would be used here")

        if isinstance(input_tensor.data_subjects, np.ndarray):
            # TODO: THIS IS A HACK FOR AA AND SHOULD __NOT__ BE MERGED INTO DEV
            # Need to convert newDSL to old DSL-> this works because for each step we only use 1 data subject at a time
            root_child = None
            while isinstance(input_tensor, PassthroughTensor):
                root_child = input_tensor.child
                input_tensor = root_child
            # input_entities = np.zeros_like(root_child)
            input_entities = input_tensor.data_subjects
        elif isinstance(input_tensor.data_subjects, DataSubjectList):
            input_entities = input_tensor.data_subjects.data_subjects_indexed
        else:
            raise NotImplementedError(
                f"Undefined behaviour for data subjects type: {type(input_tensor.data_subjects)}"
            )
        # data_subjects.data_subjects_indexed[0].reshape(-1)
        # t2 = time()
        # print("Obtained RDP Params, calculation time", t2 - t1)

        # Query budget spend of all unique data_subjects
        rdp_params = RDPParams(
            sigmas=sigmas,
            l2_norms=l2_norms,
            l2_norm_bounds=l2_norm_bounds,
            Ls=lipschitz_bounds,
            coeffs=coeffs,
        )
        # print("Finished RDP Params Initialization")
        try:
            # query and save
            mask = ledger.get_entity_overbudget_mask_for_epsilon_and_append(
                unique_entity_ids_query=input_entities,
                rdp_params=rdp_params,
                private=True,
                get_budget_for_user=get_budget_for_user,
                deduct_epsilon_for_user=deduct_epsilon_for_user,
            )
            # We had to flatten the mask so the code generalized for N-dim arrays, here we reshape it back
            reshaped_mask = mask.reshape(value.shape)
            # print("Fixed mask shape!")
            # here we have the final mask and highest possible spend has been applied
            # to the data scientists budget field in the database

            if mask is None:
                raise Exception("Failed to publish mask is None")

            # print("Obtained overbudgeted entity mask", mask.dtype)

            # multiply values by the inverted mask
            filtered_input_tensor = value * (
                1 - reshaped_mask
            )  # + gauss(0, sigma)  # Double check that noise has mean of 0

            filtered_inputs.append(filtered_input_tensor)
            # output = np.asarray(output_func(filtered_inputs) + noise)

        except Exception as e:
            # stdlib
            import traceback

            print(traceback.format_exc())
            print(f"Failed to run vectorized_publish. {e}")

    print("We have filtered all the input tensors. Now to compute the result:")

    # noise = secrets.SystemRandom().gauss(0, sigma)
    print(
        "Filtered inputs ",
        type(filtered_inputs),
        type(filtered_input_tensor),
        filtered_inputs,
    )
    # GammaTensor.convert_dsl(state_tree)
    print("Converted DSLs")
    original_output = np.asarray(output_func(filtered_inputs))
    print("original output (before noise:", original_output)
    noise = np.asarray(
        [secrets.SystemRandom().gauss(0, sigma) for _ in range(original_output.size)]
    )
    noise.resize(original_output.shape)
    print("noise: ", noise)

    output = np.asarray(output_func(filtered_inputs) + noise)

    print("got output", output, type(output), output.dtype)

    return output.squeeze()



# third party

# stdlib
from typing import Tuple
from typing import Union

# third party
import numpy as np

# relative
from ...common.serde.serializable import serializable
from ..autodp.gamma_tensor import GammaTensor
from ..autodp.phi_tensor import PhiTensor
from .utils import dp_log


@serializable(recursive_serde=True)
class Loss(object):
    """An objective function (or loss function, or optimization score
    function) is one of the two parameters required to compile a model.

    """

    __attr_allowlist__: Tuple[str, ...] = ()

    def forward(
        self,
        outputs: Union[PhiTensor, GammaTensor],
        targets: Union[PhiTensor, GammaTensor],
    ) -> Union[PhiTensor, GammaTensor]:
        """Forward function."""
        raise NotImplementedError()

    def backward(
        self,
        outputs: Union[PhiTensor, GammaTensor],
        targets: Union[PhiTensor, GammaTensor],
    ) -> Union[PhiTensor, GammaTensor]:
        """Backward function.

        Parameters
        ----------
        outputs, targets : numpy.array
            The arrays to compute the derivatives of them.

        Returns
        -------
        numpy.array
            An array of derivative.
        """
        raise NotImplementedError()

    def __str__(self) -> str:
        return self.__class__.__name__


@serializable(recursive_serde=True)
class BinaryCrossEntropy(Loss):
    __attr_allowlist__ = ("epsilon",)

    def __init__(self, epsilon: float = 1e-11):
        self.epsilon = epsilon

    def forward(
        self,
        outputs: Union[PhiTensor, GammaTensor],
        targets: Union[PhiTensor, GammaTensor],
    ) -> Union[PhiTensor, GammaTensor]:
        """Forward pass.

        .. math:: L = -t \\log(p) - (1 - t) \\log(1 - p)

        Parameters
        ----------
        outputs : numpy.array
            Predictions in (0, 1), such as sigmoidal output of a neural network.
        targets : numpy.array
            Targets in [0, 1], such as ground truth labels.
        """
        outputs = outputs.clip(self.epsilon, 1 - self.epsilon)
        log_loss = targets * dp_log(outputs) + ((targets * -1) + 1) * dp_log(
            (outputs * -1) + 1
        )
        log_loss = log_loss.sum(axis=1) * -1
        return log_loss.mean(axis=0)

    def backward(
        self,
        outputs: Union[PhiTensor, GammaTensor],
        targets: Union[PhiTensor, GammaTensor],
    ) -> Union[PhiTensor, GammaTensor]:
        """Backward pass.
        Parameters
        ----------
        outputs : numpy.array
            Predictions in (0, 1), such as sigmoidal output of a neural network.
        targets : numpy.array
            Targets in [0, 1], such as ground truth labels.
        """
        outputs = outputs.clip(self.epsilon, 1 - self.epsilon)
        divisor = outputs * ((outputs * -1) + 1)
        divisor = np.maximum(divisor.child, self.epsilon)
        return (outputs - targets) * (1.0 / divisor)



# stdlib
from typing import Dict
from typing import List
from typing import Optional

# third party
from google.protobuf.reflection import GeneratedProtocolMessageType
from typing_extensions import final

# relative
from ...... import serialize
from ......proto.grid.messages.association_messages_pb2 import (
    DeleteAssociationRequestMessage as DeleteAssociationRequestMessage_PB,
)
from ......proto.grid.messages.association_messages_pb2 import (
    GetAssociationRequestMessage as GetAssociationRequestMessage_PB,
)
from ......proto.grid.messages.association_messages_pb2 import (
    GetAssociationRequestResponse as GetAssociationRequestResponse_PB,
)
from ......proto.grid.messages.association_messages_pb2 import (
    GetAssociationRequestsMessage as GetAssociationRequestsMessage_PB,
)
from ......proto.grid.messages.association_messages_pb2 import (
    GetAssociationRequestsResponse as GetAssociationRequestsResponse_PB,
)
from ......proto.grid.messages.association_messages_pb2 import (
    ReceiveAssociationRequestMessage as ReceiveAssociationRequestMessage_PB,
)
from ......proto.grid.messages.association_messages_pb2 import (
    RespondAssociationRequestMessage as RespondAssociationRequestMessage_PB,
)
from ......proto.grid.messages.association_messages_pb2 import (
    SendAssociationRequestMessage as SendAssociationRequestMessage_PB,
)
from .....common.message import ImmediateSyftMessageWithReply
from .....common.message import ImmediateSyftMessageWithoutReply
from .....common.serde.deserialize import _deserialize
from .....common.serde.serializable import serializable
from .....common.uid import UID
from .....io.address import Address


@final
@serializable()
class SendAssociationRequestMessage(ImmediateSyftMessageWithReply):
    def __init__(
        self,
        source: str,
        target: str,
        address: Address,
        reply_to: Address,
        metadata: Dict[str, str],
        msg_id: Optional[UID] = None,
    ):
        super().__init__(address=address, msg_id=msg_id, reply_to=reply_to)
        self.source = source
        self.target = target
        self.metadata = metadata

    def _object2proto(self) -> SendAssociationRequestMessage_PB:
        """Returns a protobuf serialization of self.
        As a requirement of all objects which inherit from Serializable,
        this method transforms the current object into the corresponding
        Protobuf object so that it can be further serialized.
        :return: returns a protobuf object
        :rtype: SendAssociationRequestMessage_PB
        .. note::
            This method is purely an internal method. Please use serialize(object) or one of
            the other public serialization methods if you wish to serialize an
            object.
        """
        return SendAssociationRequestMessage_PB(
            msg_id=serialize(self.id),
            address=serialize(self.address),
            source=self.source,
            target=self.target,
            reply_to=serialize(self.reply_to),
            metadata=self.metadata,
        )

    @staticmethod
    def _proto2object(
        proto: SendAssociationRequestMessage_PB,
    ) -> "SendAssociationRequestMessage":
        """Creates a SendAssociationRequestMessage from a protobuf
        As a requirement of all objects which inherit from Serializable,
        this method transforms a protobuf object into an instance of this class.
        :return: returns an instance of SignalingOfferMessage
        :rtype: SendAssociationRequestMessage
        .. note::
            This method is purely an internal method. Please use syft.deserialize()
            if you wish to deserialize an object.
        """

        return SendAssociationRequestMessage(
            msg_id=_deserialize(blob=proto.msg_id),
            address=_deserialize(blob=proto.address),
            source=proto.source,
            target=proto.target,
            reply_to=_deserialize(blob=proto.reply_to),
            metadata=dict(proto.metadata),
        )

    @staticmethod
    def get_protobuf_schema() -> GeneratedProtocolMessageType:
        """Return the type of protobuf object which stores a class of this type
        As a part of serialization and deserialization, we need the ability to
        lookup the protobuf object type directly from the object type. This
        static method allows us to do this.
        Importantly, this method is also used to create the reverse lookup ability within
        the metaclass of Serializable. In the metaclass, it calls this method and then
        it takes whatever type is returned from this method and adds an attribute to it
        with the type of this class attached to it. See the MetaSerializable class for
        details.
        :return: the type of protobuf object which corresponds to this class.
        :rtype: GeneratedProtocolMessageType
        """

        return SendAssociationRequestMessage_PB


@final
@serializable()
class ReceiveAssociationRequestMessage(ImmediateSyftMessageWithoutReply):
    def __init__(
        self,
        address: Address,
        source: str,
        target: str,
        metadata: Dict[str, str],
        msg_id: Optional[UID] = None,
        response: Optional[str] = "",
    ):
        super().__init__(address=address, msg_id=msg_id)
        self.metadata = metadata
        self.response = response
        self.source = source
        self.target = target

    def _object2proto(self) -> ReceiveAssociationRequestMessage_PB:
        """Returns a protobuf serialization of self.
        As a requirement of all objects which inherit from Serializable,
        this method transforms the current object into the corresponding
        Protobuf object so that it can be further serialized.
        :return: returns a protobuf object
        :rtype: ReceiveAssociationRequestMessage_PB
        .. note::
            This method is purely an internal method. Please use serialize(object) or one of
            the other public serialization methods if you wish to serialize an
            object.
        """

        return ReceiveAssociationRequestMessage_PB(
            msg_id=serialize(self.id),
            address=serialize(self.address),
            response=self.response,
            metadata=self.metadata,
            source=self.source,
            target=self.target,
        )

    @staticmethod
    def _proto2object(
        proto: ReceiveAssociationRequestMessage_PB,
    ) -> "ReceiveAssociationRequestMessage":
        """Creates a ReceiveAssociationRequestMessage from a protobuf
        As a requirement of all objects which inherit from Serializable,
        this method transforms a protobuf object into an instance of this class.
        :return: returns an instance of SignalingOfferMessage
        :rtype: ReceiveAssociationRequestMessage
        .. note::
            This method is purely an internal method. Please use syft.deserialize()
            if you wish to deserialize an object.
        """

        return ReceiveAssociationRequestMessage(
            msg_id=_deserialize(blob=proto.msg_id),
            address=_deserialize(blob=proto.address),
            response=proto.response,
            metadata=proto.metadata,
            source=proto.source,
            target=proto.target,
        )

    @staticmethod
    def get_protobuf_schema() -> GeneratedProtocolMessageType:
        """Return the type of protobuf object which stores a class of this type
        As a part of serialization and deserialization, we need the ability to
        lookup the protobuf object type directly from the object type. This
        static method allows us to do this.
        Importantly, this method is also used to create the reverse lookup ability within
        the metaclass of Serializable. In the metaclass, it calls this method and then
        it takes whatever type is returned from this method and adds an attribute to it
        with the type of this class attached to it. See the MetaSerializable class for
        details.
        :return: the type of protobuf object which corresponds to this class.
        :rtype: GeneratedProtocolMessageType
        """

        return ReceiveAssociationRequestMessage_PB


@final
@serializable()
class RespondAssociationRequestMessage(ImmediateSyftMessageWithReply):
    def __init__(
        self,
        address: Address,
        response: str,
        reply_to: Address,
        source: str,
        target: str,
        msg_id: Optional[UID] = None,
    ):
        super().__init__(address=address, msg_id=msg_id, reply_to=reply_to)
        self.response = response
        self.source = source
        self.target = target

    def _object2proto(self) -> RespondAssociationRequestMessage_PB:
        """Returns a protobuf serialization of self.
        As a requirement of all objects which inherit from Serializable,
        this method transforms the current object into the corresponding
        Protobuf object so that it can be further serialized.
        :return: returns a protobuf object
        :rtype: RespondAssociationRequestMessage_PB
        .. note::
            This method is purely an internal method. Please use serialize(object) or one of
            the other public serialization methods if you wish to serialize an
            object.
        """
        return RespondAssociationRequestMessage_PB(
            msg_id=serialize(self.id),
            address=serialize(self.address),
            source=self.source,
            target=self.target,
            response=self.response,
            reply_to=serialize(self.reply_to),
        )

    @staticmethod
    def _proto2object(
        proto: RespondAssociationRequestMessage_PB,
    ) -> "RespondAssociationRequestMessage":
        """Creates a RespondAssociationRequestMessage from a protobuf
        As a requirement of all objects which inherit from Serializable,
        this method transforms a protobuf object into an instance of this class.
        :return: returns an instance of SignalingOfferMessage
        :rtype: RespondAssociationRequestMessage
        .. note::
            This method is purely an internal method. Please use syft.deserialize()
            if you wish to deserialize an object.
        """

        return RespondAssociationRequestMessage(
            msg_id=_deserialize(blob=proto.msg_id),
            address=_deserialize(blob=proto.address),
            source=proto.source,
            target=proto.target,
            response=proto.response,
            reply_to=_deserialize(blob=proto.reply_to),
        )

    @staticmethod
    def get_protobuf_schema() -> GeneratedProtocolMessageType:
        """Return the type of protobuf object which stores a class of this type
        As a part of serialization and deserialization, we need the ability to
        lookup the protobuf object type directly from the object type. This
        static method allows us to do this.
        Importantly, this method is also used to create the reverse lookup ability within
        the metaclass of Serializable. In the metaclass, it calls this method and then
        it takes whatever type is returned from this method and adds an attribute to it
        with the type of this class attached to it. See the MetaSerializable class for
        details.
        :return: the type of protobuf object which corresponds to this class.
        :rtype: GeneratedProtocolMessageType
        """

        return RespondAssociationRequestMessage_PB


@final
@serializable()
class GetAssociationRequestMessage(ImmediateSyftMessageWithReply):
    def __init__(
        self,
        address: Address,
        association_id: int,
        reply_to: Address,
        msg_id: Optional[UID] = None,
    ):
        super().__init__(address=address, msg_id=msg_id, reply_to=reply_to)
        self.association_id = association_id

    def _object2proto(self) -> GetAssociationRequestMessage_PB:
        """Returns a protobuf serialization of self.
        As a requirement of all objects which inherit from Serializable,
        this method transforms the current object into the corresponding
        Protobuf object so that it can be further serialized.
        :return: returns a protobuf object
        :rtype: GetAssociationRequestMessage_PB
        .. note::
            This method is purely an internal method. Please use serialize(object) or one of
            the other public serialization methods if you wish to serialize an
            object.
        """
        return GetAssociationRequestMessage_PB(
            msg_id=serialize(self.id),
            address=serialize(self.address),
            association_id=self.association_id,
            reply_to=serialize(self.reply_to),
        )

    @staticmethod
    def _proto2object(
        proto: GetAssociationRequestMessage_PB,
    ) -> "GetAssociationRequestMessage":
        """Creates a GetAssociationRequestMessage from a protobuf
        As a requirement of all objects which inherit from Serializable,
        this method transforms a protobuf object into an instance of this class.
        :return: returns an instance of SignalingOfferMessage
        :rtype: GetAssociationRequestMessage
        .. note::
            This method is purely an internal method. Please use syft.deserialize()
            if you wish to deserialize an object.
        """

        return GetAssociationRequestMessage(
            msg_id=_deserialize(blob=proto.msg_id),
            address=_deserialize(blob=proto.address),
            association_id=proto.association_id,
            reply_to=_deserialize(blob=proto.reply_to),
        )

    @staticmethod
    def get_protobuf_schema() -> GeneratedProtocolMessageType:
        """Return the type of protobuf object which stores a class of this type
        As a part of serialization and deserialization, we need the ability to
        lookup the protobuf object type directly from the object type. This
        static method allows us to do this.
        Importantly, this method is also used to create the reverse lookup ability within
        the metaclass of Serializable. In the metaclass, it calls this method and then
        it takes whatever type is returned from this method and adds an attribute to it
        with the type of this class attached to it. See the MetaSerializable class for
        details.
        :return: the type of protobuf object which corresponds to this class.
        :rtype: GeneratedProtocolMessageType
        """

        return GetAssociationRequestMessage_PB


@final
@serializable()
class GetAssociationRequestResponse(ImmediateSyftMessageWithoutReply):
    def __init__(
        self,
        address: Address,
        content: Dict,
        source: str,
        target: str,
        msg_id: Optional[UID] = None,
    ):
        super().__init__(address=address, msg_id=msg_id)
        self.content = content
        self.source = source
        self.target = target

    def _object2proto(self) -> GetAssociationRequestResponse_PB:
        """Returns a protobuf serialization of self.
        As a requirement of all objects which inherit from Serializable,
        this method transforms the current object into the corresponding
        Protobuf object so that it can be further serialized.
        :return: returns a protobuf object
        :rtype: SignalingOfferMessage_PB
        .. note::
            This method is purely an internal method. Please use serialize(object) or one of
            the other public serialization methods if you wish to serialize an
            object.
        """
        return GetAssociationRequestResponse_PB(
            msg_id=serialize(self.id),
            address=serialize(self.address),
            content=self.content,
            source=self.source,
            target=self.target,
        )

    @staticmethod
    def _proto2object(
        proto: GetAssociationRequestResponse_PB,
    ) -> "GetAssociationRequestResponse":
        """Creates a SignalingOfferMessage from a protobuf
        As a requirement of all objects which inherit from Serializable,
        this method transforms a protobuf object into an instance of this class.
        :return: returns an instance of SignalingOfferMessage
        :rtype: SignalingOfferMessage
        .. note::
            This method is purely an internal method. Please use syft.deserialize()
            if you wish to deserialize an object.
        """

        return GetAssociationRequestResponse(
            msg_id=_deserialize(blob=proto.msg_id),
            address=_deserialize(blob=proto.address),
            content=dict(proto.content),
            source=proto.source,
            target=proto.target,
        )

    @staticmethod
    def get_protobuf_schema() -> GeneratedProtocolMessageType:
        """Return the type of protobuf object which stores a class of this type
        As a part of serialization and deserialization, we need the ability to
        lookup the protobuf object type directly from the object type. This
        static method allows us to do this.
        Importantly, this method is also used to create the reverse lookup ability within
        the metaclass of Serializable. In the metaclass, it calls this method and then
        it takes whatever type is returned from this method and adds an attribute to it
        with the type of this class attached to it. See the MetaSerializable class for
        details.
        :return: the type of protobuf object which corresponds to this class.
        :rtype: GeneratedProtocolMessageType
        """

        return GetAssociationRequestResponse_PB


@final
@serializable()
class GetAssociationRequestsMessage(ImmediateSyftMessageWithReply):
    def __init__(
        self,
        address: Address,
        reply_to: Address,
        msg_id: Optional[UID] = None,
    ):
        super().__init__(address=address, msg_id=msg_id, reply_to=reply_to)

    def _object2proto(self) -> GetAssociationRequestsMessage_PB:
        """Returns a protobuf serialization of self.
        As a requirement of all objects which inherit from Serializable,
        this method transforms the current object into the corresponding
        Protobuf object so that it can be further serialized.
        :return: returns a protobuf object
        :rtype: GetAssociationRequestsMessage_PB
        .. note::
            This method is purely an internal method. Please use serialize(object) or one of
            the other public serialization methods if you wish to serialize an
            object.
        """
        return GetAssociationRequestsMessage_PB(
            msg_id=serialize(self.id),
            address=serialize(self.address),
            reply_to=serialize(self.reply_to),
        )

    @staticmethod
    def _proto2object(
        proto: GetAssociationRequestsMessage_PB,
    ) -> "GetAssociationRequestsMessage":
        """Creates a GetAssociationRequestsMessage from a protobuf
        As a requirement of all objects which inherit from Serializable,
        this method transforms a protobuf object into an instance of this class.
        :return: returns an instance of SignalingOfferMessage
        :rtype: GetAssociationRequestsMessage
        .. note::
            This method is purely an internal method. Please use syft.deserialize()
            if you wish to deserialize an object.
        """

        return GetAssociationRequestsMessage(
            msg_id=_deserialize(blob=proto.msg_id),
            address=_deserialize(blob=proto.address),
            reply_to=_deserialize(blob=proto.reply_to),
        )

    @staticmethod
    def get_protobuf_schema() -> GeneratedProtocolMessageType:
        """Return the type of protobuf object which stores a class of this type
        As a part of serialization and deserialization, we need the ability to
        lookup the protobuf object type directly from the object type. This
        static method allows us to do this.
        Importantly, this method is also used to create the reverse lookup ability within
        the metaclass of Serializable. In the metaclass, it calls this method and then
        it takes whatever type is returned from this method and adds an attribute to it
        with the type of this class attached to it. See the MetaSerializable class for
        details.
        :return: the type of protobuf object which corresponds to this class.
        :rtype: GeneratedProtocolMessageType
        """

        return GetAssociationRequestsMessage_PB


@final
@serializable()
class GetAssociationRequestsResponse(ImmediateSyftMessageWithoutReply):
    def __init__(
        self,
        address: Address,
        content: List[Dict],
        msg_id: Optional[UID] = None,
    ):
        super().__init__(address=address, msg_id=msg_id)
        self.content = content

    def _object2proto(self) -> GetAssociationRequestsResponse_PB:
        """Returns a protobuf serialization of self.
        As a requirement of all objects which inherit from Serializable,
        this method transforms the current object into the corresponding
        Protobuf object so that it can be further serialized.
        :return: returns a protobuf object
        :rtype: SignalingOfferMessage_PB
        .. note::
            This method is purely an internal method. Please use serialize(object) or one of
            the other public serialization methods if you wish to serialize an
            object.
        """
        msg = GetAssociationRequestsResponse_PB(
            msg_id=serialize(self.id),
            address=serialize(self.address),
        )

        metadata_constructor = GetAssociationRequestsResponse_PB.metadata_container
        _ = [
            msg.content.append(metadata_constructor(metadata=metadata))
            for metadata in self.content
        ]
        return msg

    @staticmethod
    def _proto2object(
        proto: GetAssociationRequestsResponse_PB,
    ) -> "GetAssociationRequestsResponse":
        """Creates a SignalingOfferMessage from a protobuf
        As a requirement of all objects which inherit from Serializable,
        this method transforms a protobuf object into an instance of this class.
        :return: returns an instance of SignalingOfferMessage
        :rtype: SignalingOfferMessage
        .. note::
            This method is purely an internal method. Please use syft.deserialize()
            if you wish to deserialize an object.
        """
        return GetAssociationRequestsResponse(
            msg_id=_deserialize(blob=proto.msg_id),
            address=_deserialize(blob=proto.address),
            content=[
                dict(metadata_container.metadata)
                for metadata_container in proto.content
            ],
        )

    @staticmethod
    def get_protobuf_schema() -> GeneratedProtocolMessageType:
        """Return the type of protobuf object which stores a class of this type
        As a part of serialization and deserialization, we need the ability to
        lookup the protobuf object type directly from the object type. This
        static method allows us to do this.
        Importantly, this method is also used to create the reverse lookup ability within
        the metaclass of Serializable. In the metaclass, it calls this method and then
        it takes whatever type is returned from this method and adds an attribute to it
        with the type of this class attached to it. See the MetaSerializable class for
        details.
        :return: the type of protobuf object which corresponds to this class.
        :rtype: GeneratedProtocolMessageType
        """

        return GetAssociationRequestsResponse_PB


@serializable()
@final
class DeleteAssociationRequestMessage(ImmediateSyftMessageWithReply):
    def __init__(
        self,
        address: Address,
        association_id: int,
        reply_to: Address,
        msg_id: Optional[UID] = None,
    ):
        super().__init__(address=address, msg_id=msg_id, reply_to=reply_to)
        self.association_id = association_id

    def _object2proto(self) -> DeleteAssociationRequestMessage_PB:
        """Returns a protobuf serialization of self.
        As a requirement of all objects which inherit from Serializable,
        this method transforms the current object into the corresponding
        Protobuf object so that it can be further serialized.
        :return: returns a protobuf object
        :rtype: DeleteAssociationRequestMessage_PB
        .. note::
            This method is purely an internal method. Please use serialize(object) or one of
            the other public serialization methods if you wish to serialize an
            object.
        """
        return DeleteAssociationRequestMessage_PB(
            msg_id=serialize(self.id),
            address=serialize(self.address),
            association_id=self.association_id,
            reply_to=serialize(self.reply_to),
        )

    @staticmethod
    def _proto2object(
        proto: DeleteAssociationRequestMessage_PB,
    ) -> "DeleteAssociationRequestMessage":
        """Creates a DeleteAssociationRequestMessage from a protobuf
        As a requirement of all objects which inherit from Serializable,
        this method transforms a protobuf object into an instance of this class.
        :return: returns an instance of SignalingOfferMessage
        :rtype: DeleteAssociationRequestMessage
        .. note::
            This method is purely an internal method. Please use syft.deserialize()
            if you wish to deserialize an object.
        """

        return DeleteAssociationRequestMessage(
            msg_id=_deserialize(blob=proto.msg_id),
            address=_deserialize(blob=proto.address),
            association_id=proto.association_id,
            reply_to=_deserialize(blob=proto.reply_to),
        )

    @staticmethod
    def get_protobuf_schema() -> GeneratedProtocolMessageType:
        """Return the type of protobuf object which stores a class of this type
        As a part of serialization and deserialization, we need the ability to
        lookup the protobuf object type directly from the object type. This
        static method allows us to do this.
        Importantly, this method is also used to create the reverse lookup ability within
        the metaclass of Serializable. In the metaclass, it calls this method and then
        it takes whatever type is returned from this method and adds an attribute to it
        with the type of this class attached to it. See the MetaSerializable class for
        details.
        :return: the type of protobuf object which corresponds to this class.
        :rtype: GeneratedProtocolMessageType
        """

        return DeleteAssociationRequestMessage_PB



# future
from __future__ import annotations

# stdlib
import operator
import secrets
from typing import Any
from typing import Callable
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple
from typing import Union

# third party
import numpy as np
import pandas as pd
import torch as th

# syft absolute
import syft as sy

# relative
from ... import lib
from ...ast.klass import pointerize_args_and_kwargs
from ...core.adp.data_subject_ledger import DataSubjectLedger
from ...util import inherit_tags
from ..common.serde.capnp import CapnpModule
from ..common.serde.capnp import chunk_bytes
from ..common.serde.capnp import combine_bytes
from ..common.serde.capnp import get_capnp_schema
from ..common.serde.capnp import serde_magic_header
from ..common.serde.serializable import serializable
from ..common.uid import UID
from ..node.abstract.node import AbstractNodeClient
from ..node.common.action import smpc_action_functions
from ..node.common.action.run_class_method_smpc_action import RunClassMethodSMPCAction

# from ..node.domain.client import DomainClient
from ..pointer.pointer import Pointer
from .ancestors import PhiTensorAncestor
from .autodp.gamma_tensor import GammaTensor
from .autodp.gamma_tensor import TensorWrappedGammaTensorPointer
from .autodp.phi_tensor import PhiTensor
from .autodp.phi_tensor import TensorWrappedPhiTensorPointer
from .config import DEFAULT_FLOAT_NUMPY_TYPE
from .config import DEFAULT_INT_NUMPY_TYPE
from .fixed_precision_tensor_ancestor import FixedPrecisionTensorAncestor
from .passthrough import PassthroughTensor  # type: ignore
from .smpc import context
from .smpc import utils
from .smpc.mpc_tensor import MPCTensor


class TensorPointer(Pointer):

    # Must set these at class init time despite
    # the fact that klass.Class tries to override them (unsuccessfully)
    __name__ = "TensorPointer"
    __module__ = "syft.core.tensor.tensor"

    def __init__(
        self,
        client: Any,
        id_at_location: Optional[UID] = None,
        object_type: str = "",
        tags: Optional[List[str]] = None,
        description: str = "",
        public_shape: Optional[Tuple[int, ...]] = None,
        public_dtype: Optional[Union[str, np.dtype]] = None,
    ):

        super().__init__(
            client=client,
            id_at_location=id_at_location,
            object_type=object_type,
            tags=tags,
            description=description,
        )

        self.public_shape = public_shape

        if isinstance(public_dtype, str):
            self.public_dtype = np.dtype(public_dtype)
        else:
            self.public_dtype = public_dtype

    def share(self, *parties: Tuple[AbstractNodeClient, ...]) -> MPCTensor:
        all_parties = list(parties) + [self.client]
        ring_size = utils.TYPE_TO_RING_SIZE.get(self.public_dtype, None)
        self_mpc = MPCTensor(
            secret=self,
            shape=self.public_shape,
            parties=all_parties,
            ring_size=ring_size,
        )
        return self_mpc

    def _apply_tensor_op(
        self,
        other: Any,
        op_str: str,
        *args: Tuple[Any, ...],
        **kwargs: Any,
    ) -> Any:
        # we want to get the return type which matches the attr_path_and_name
        # so we ask lib_ast for the return type name that matches out
        # attr_path_and_name and then use that to get the actual pointer klass
        # then set the result to that pointer klass

        op = f"__{op_str}__" if op_str != "concatenate" else "concatenate"
        # remove this to dunder method before merge.
        attr_path_and_name = f"syft.core.tensor.tensor.Tensor.{op}"
        seed_id_locations = kwargs.pop("seed_id_locations", None)
        if seed_id_locations is None:
            seed_id_locations = secrets.randbits(64)

        id_at_location = smpc_action_functions.get_id_at_location_from_op(
            seed_id_locations, op
        )

        # QUESTION can the id_at_location be None?
        result_id_at_location = id_at_location

        result = TensorPointer(client=self.client)
        result.id_at_location = result_id_at_location

        if result_id_at_location is not None:
            # first downcast anything primitive which is not already PyPrimitive
            (
                downcast_args,
                downcast_kwargs,
            ) = lib.python.util.downcast_args_and_kwargs(
                args=[self, other], kwargs=kwargs
            )

            # then we convert anything which isnt a pointer into a pointer
            pointer_args, pointer_kwargs = pointerize_args_and_kwargs(
                args=downcast_args,
                kwargs=downcast_kwargs,
                client=self.client,
                gc_enabled=False,
            )

            cmd = RunClassMethodSMPCAction(
                path=attr_path_and_name,
                _self=self,
                args=pointer_args,
                kwargs=pointer_kwargs,
                id_at_location=result_id_at_location,
                seed_id_locations=seed_id_locations,
                address=self.client.address,
            )
            self.client.send_immediate_msg_without_reply(msg=cmd)

        inherit_tags(
            attr_path_and_name=attr_path_and_name,
            result=result,
            self_obj=self,
            args=[other],
            kwargs={},
        )

        result_public_shape = None
        result_public_dtype = None

        if isinstance(other, TensorPointer):
            other_shape = other.public_shape
            other_dtype = other.public_dtype
        elif isinstance(other, (int, float)):
            other_shape = (1,)
            other_dtype = np.int32
        elif isinstance(other, bool):
            other_shape = (1,)
            other_dtype = np.dtype("bool")
        elif isinstance(other, np.ndarray):
            other_shape = other.shape
            other_dtype = other.dtype

        else:
            raise ValueError(f"Invalid Type for TensorPointer:{type(other)}")

        if self.public_shape is not None and other_shape is not None:
            result_public_shape = utils.get_shape(
                op_str, self.public_shape, other_shape
            )

        if self.public_dtype is not None and other_dtype is not None:
            if self.public_dtype != other_dtype:
                raise ValueError(
                    f"Type for self and other do not match ({self.public_dtype} vs {other_dtype})"
                )
            result_public_dtype = self.public_dtype

        result.public_shape = result_public_shape
        result.public_dtype = result_public_dtype

        return result

    @staticmethod
    def _apply_op(
        self: TensorPointer,
        other: Union[TensorPointer, MPCTensor, int, float, np.ndarray],
        op_str: str,
        **kwargs: Dict[str, Any],
    ) -> Union[MPCTensor, TensorPointer]:
        """Performs the operation based on op_str

        Args:
            other (Union[TensorPointer,MPCTensor,int,float,np.ndarray]): second operand.

        Returns:
            Tuple[MPCTensor,Union[MPCTensor,int,float,np.ndarray]] : Result of the operation
        """

        if isinstance(other, TensorPointer) and self.client != other.client:

            parties = [self.client, other.client]
            self_mpc = MPCTensor(secret=self, shape=self.public_shape, parties=parties)
            other_mpc = MPCTensor(
                secret=other, shape=other.public_shape, parties=parties
            )
            if op_str != "concatenate":
                op = getattr(operator, op_str)
                return op(self_mpc, other_mpc)
            else:
                return self_mpc.concatenate(other_mpc)

        elif isinstance(other, MPCTensor):
            if op_str != "concatenate":
                op = getattr(operator, op_str)
                return op(other, self)
            else:
                return other.concatenate(self)

        return self._apply_tensor_op(other=other, op_str=op_str, **kwargs)

    def __add__(
        self,
        other: Union[TensorPointer, MPCTensor, int, float, np.ndarray],
        **kwargs: Dict[str, Any],
    ) -> Union[TensorPointer, MPCTensor]:
        """Apply the "add" operation between "self" and "other"

        Args:
            y (Union[TensorPointer,MPCTensor,int,float,np.ndarray]) : second operand.

        Returns:
            Union[TensorPointer,MPCTensor] : Result of the operation.
        """
        return TensorPointer._apply_op(self, other, "add", **kwargs)

    def __sub__(
        self,
        other: Union[TensorPointer, MPCTensor, int, float, np.ndarray],
        **kwargs: Dict[str, Any],
    ) -> Union[TensorPointer, MPCTensor]:
        """Apply the "sub" operation between "self" and "other"

        Args:
            y (Union[TensorPointer,MPCTensor,int,float,np.ndarray]) : second operand.

        Returns:
            Union[TensorPointer,MPCTensor] : Result of the operation.
        """
        return TensorPointer._apply_op(self, other, "sub", **kwargs)

    def __mul__(
        self,
        other: Union[TensorPointer, MPCTensor, int, float, np.ndarray],
        **kwargs: Dict[str, Any],
    ) -> Union[TensorPointer, MPCTensor]:
        """Apply the "mul" operation between "self" and "other"

        Args:
            y (Union[TensorPointer,MPCTensor,int,float,np.ndarray]) : second operand.

        Returns:
            Union[TensorPointer,MPCTensor] : Result of the operation.
        """
        return TensorPointer._apply_op(self, other, "mul", **kwargs)

    def __matmul__(
        self,
        other: Union[TensorPointer, MPCTensor, int, float, np.ndarray],
        **kwargs: Dict[str, Any],
    ) -> Union[TensorPointer, MPCTensor]:
        """Apply the "matmul" operation between "self" and "other"

        Args:
            y (Union[TensorPointer,MPCTensor,int,float,np.ndarray]) : second operand.

        Returns:
            Union[TensorPointer,MPCTensor] : Result of the operation.
        """
        return TensorPointer._apply_op(self, other, "matmul", **kwargs)

    def __truediv__(
        self,
        other: Union[TensorPointer, MPCTensor, int, float, np.ndarray],
        **kwargs: Dict[str, Any],
    ) -> Union[TensorPointer, MPCTensor]:
        """Apply the "mul" operation between "self" and "other"

        Args:
            y (Union[TensorPointer,MPCTensor,int,float,np.ndarray]) : second operand.

        Returns:
            Union[TensorPointer,MPCTensor] : Result of the operation.
        """
        return TensorPointer._apply_op(self, other, "truediv", **kwargs)

    def __rtruediv__(
        self,
        other: Union[TensorPointer, MPCTensor, int, float, np.ndarray],
        **kwargs: Dict[str, Any],
    ) -> Union[TensorPointer, MPCTensor]:
        """Apply the "mul" operation between "self" and "other"

        Args:
            y (Union[TensorPointer,MPCTensor,int,float,np.ndarray]) : second operand.

        Returns:
            Union[TensorPointer,MPCTensor] : Result of the operation.
        """
        raise NotImplementedError

    def __lt__(
        self, other: Union[TensorPointer, MPCTensor, int, float, np.ndarray]
    ) -> Union[TensorPointer, MPCTensor]:
        """Apply the "lt" operation between "self" and "other"

        Args:
            y (Union[TensorPointer,MPCTensor,int,float,np.ndarray]) : second operand.

        Returns:
            Union[TensorPointer,MPCTensor] : Result of the operation.
        """

        return TensorPointer._apply_op(self, other, "lt")

    def __gt__(
        self, other: Union[TensorPointer, MPCTensor, int, float, np.ndarray]
    ) -> Union[TensorPointer, MPCTensor]:
        """Apply the "gt" operation between "self" and "other"

        Args:
            y (Union[TensorPointer,MPCTensor,int,float,np.ndarray]) : second operand.

        Returns:
            Union[TensorPointer,MPCTensor] : Result of the operation.
        """

        return TensorPointer._apply_op(self, other, "gt")

    def __ge__(
        self, other: Union[TensorPointer, MPCTensor, int, float, np.ndarray]
    ) -> Union[TensorPointer, MPCTensor]:
        """Apply the "ge" operation between "self" and "other"

        Args:
            y (Union[TensorPointer,MPCTensor,int,float,np.ndarray]) : second operand.

        Returns:
            Union[TensorPointer,MPCTensor] : Result of the operation.
        """

        return TensorPointer._apply_op(self, other, "ge")

    def __le__(
        self, other: Union[TensorPointer, MPCTensor, int, float, np.ndarray]
    ) -> Union[TensorPointer, MPCTensor]:
        """Apply the "le" operation between "self" and "other"

        Args:
            y (Union[TensorPointer,MPCTensor,int,float,np.ndarray]) : second operand.

        Returns:
            Union[TensorPointer,MPCTensor] : Result of the operation.
        """

        return TensorPointer._apply_op(self, other, "le")

    def __eq__(  # type: ignore
        self, other: Union[TensorPointer, MPCTensor, int, float, np.ndarray]
    ) -> Union[TensorPointer, MPCTensor]:
        """Apply the "eq" operation between "self" and "other"

        Args:
            y (Union[TensorPointer,MPCTensor,int,float,np.ndarray]) : second operand.

        Returns:
            Union[TensorPointer,MPCTensor] : Result of the operation.
        """

        return TensorPointer._apply_op(self, other, "eq")

    def __ne__(  # type: ignore
        self, other: Union[TensorPointer, MPCTensor, int, float, np.ndarray]
    ) -> Union[TensorPointer, MPCTensor]:
        """Apply the "ne" operation between "self" and "other"

        Args:
            y (Union[TensorPointer,MPCTensor,int,float,np.ndarray]) : second operand.

        Returns:
            Union[TensorPointer,MPCTensor] : Result of the operation.
        """

        return TensorPointer._apply_op(self, other, "ne")

    def concatenate(
        self,
        other: TensorPointer,
        *args: List[Any],
        **kwargs: Dict[str, Any],
    ) -> Union[TensorPointer, MPCTensor]:
        """Apply the "add" operation between "self" and "other"

        Args:
            y (Union[TensorPointer,MPCTensor,int,float,np.ndarray]) : second operand.

        Returns:
            Union[TensorPointer,MPCTensor] : Result of the operation.
        """
        return TensorPointer._apply_op(self, other, "concatenate")


def to32bit(np_array: np.ndarray, verbose: bool = True) -> np.ndarray:

    if np_array.dtype == np.int64:
        if verbose:
            print("Casting internal tensor to int32")
        out = np_array.astype(np.int32)

    elif np_array.dtype == np.float64:

        if verbose:
            print("Casting internal tensor to float32")
        out = np_array.astype(np.float32)

    else:
        out = np_array

    return out


@serializable(capnp_bytes=True)
class Tensor(
    PassthroughTensor,
    PhiTensorAncestor,
    FixedPrecisionTensorAncestor,
    # MPCTensorAncestor,
):

    # __attr_allowlist__ = ["child", "tag_name", "public_shape", "public_dtype"]

    PointerClassOverride = TensorPointer

    def __init__(
        self,
        child: Any,
        public_shape: Optional[Tuple[int, ...]] = None,
        public_dtype: Optional[str] = None,
    ) -> None:
        """data must be a list of numpy array"""

        if isinstance(child, list) or np.isscalar(child):
            child = np.array(child)

        if isinstance(child, th.Tensor):
            print(
                "Converting PyTorch tensor to numpy tensor for internal representation..."
            )
            child = child.numpy()

        # Added for convenience- might need to double check if dtype changes?
        if isinstance(child, pd.Series):
            child = child.to_numpy()

        if (
            not isinstance(child, PassthroughTensor)
            and not isinstance(child, np.ndarray)
            and not isinstance(child, GammaTensor)
        ):

            raise Exception(
                f"Data: {child} ,type: {type(child)} must be list or nd.array "
            )

        if not isinstance(child, (np.ndarray, PassthroughTensor, GammaTensor)) or (
            getattr(child, "dtype", None)
            not in [DEFAULT_INT_NUMPY_TYPE, DEFAULT_FLOAT_NUMPY_TYPE, np.bool]
            and getattr(child, "dtype", None) is not None
        ):
            raise TypeError(
                "You tried to pass an a tensor of type:"
                + str(type(child))
                + " with child.dtype == "
                + str(getattr(child, "dtype", None))
                + ". Syft tensor objects only supports numpy objects of "
                + f"{DEFAULT_INT_NUMPY_TYPE,DEFAULT_FLOAT_NUMPY_TYPE,np.bool_}. "
                + "Please pass in either the supported types or change the default types in syft/core/tensor/config.py "
            )

        kwargs = {"child": child}
        super().__init__(**kwargs)

        # set public shape to be the shape of the data since we have access to it at present
        if public_shape is None:
            public_shape = tuple(self.shape)

        # set public dtype to be the dtype of the data since we have access to it at present
        if public_dtype is None:
            public_dtype = str(self.dtype)

        self.tag_name: str = ""
        self.public_shape = public_shape
        self.public_dtype = public_dtype

    def tag(self, name: str) -> Tensor:
        self.tag_name = name
        return self

    def exp(self) -> Tensor:
        if hasattr(self.child, "exp"):
            return self.__class__(self.child.exp())
        else:
            raise ValueError("Tensor Chain does not have exp function")

    def reciprocal(self) -> Tensor:
        if hasattr(self.child, "reciprocal"):
            return self.__class__(self.child.reciprocal())
        else:
            raise ValueError("Tensor Chain does not have reciprocal function")

    def softmax(self) -> Tensor:
        if hasattr(self.child, "softmax"):
            return self.__class__(self.child.softmax())
        else:
            raise ValueError("Tensor Chain does not have softmax function")

    def one_hot(self) -> Tensor:
        if hasattr(self.child, "one_hot"):
            return self.__class__(self.child.one_hot())
        else:
            raise ValueError("Tensor Chain does not have one_hot function")

    @property
    def shape(self) -> Tuple[Any, ...]:
        try:
            return self.child.shape
        except Exception:  # nosec
            return self.public_shape

    @property
    def proxy_public_kwargs(self) -> Dict[str, Any]:
        return {"public_shape": self.public_shape, "public_dtype": self.public_dtype}

    def init_pointer(
        self,
        client: Any,
        id_at_location: Optional[UID] = None,
        object_type: str = "",
        tags: Optional[List[str]] = None,
        description: str = "",
    ) -> Pointer:
        # relative

        if isinstance(self.child, PhiTensor):
            return TensorWrappedPhiTensorPointer(
                data_subjects=self.child.data_subjects,
                client=client,
                id_at_location=id_at_location,
                object_type=object_type,
                tags=tags,
                description=description,
                min_vals=self.child.min_vals,
                max_vals=self.child.max_vals,
                public_shape=getattr(self, "public_shape", None),
                public_dtype=getattr(self, "public_dtype", None),
            )
        elif isinstance(self.child, GammaTensor):
            return TensorWrappedGammaTensorPointer(
                data_subjects=self.child.data_subjects,
                client=client,
                id_at_location=id_at_location,
                object_type=object_type,
                tags=tags,
                description=description,
                min_vals=self.child.min_vals,
                max_vals=self.child.max_vals,
                public_shape=getattr(self, "public_shape", None),
                public_dtype=getattr(self, "public_dtype", None),
            )
        else:
            return TensorPointer(
                client=client,
                id_at_location=id_at_location,
                object_type=object_type,
                tags=tags,
                description=description,
                public_shape=getattr(self, "public_shape", None),
                public_dtype=getattr(self, "public_dtype", None),
            )

    def publish(
        self,
        get_budget_for_user: Callable,
        deduct_epsilon_for_user: Callable,
        ledger: DataSubjectLedger,
        sigma: float,
    ) -> Any:
        return self.child.publish(
            get_budget_for_user, deduct_epsilon_for_user, ledger, sigma
        )

    # TODO: remove after moving private compare to sharetensor level
    def bit_decomposition(self, ring_size: Union[int, str], bitwise: bool) -> None:
        context.tensor_values = self
        if isinstance(self.child, PhiTensor):
            self.child.child.child.bit_decomposition(ring_size, bitwise)
        else:
            self.child.bit_decomposition(ring_size, bitwise)

        return None

    def mpc_swap(self, other: Tensor) -> Tensor:
        self.child.child = other.child.child
        return self

    def _object2bytes(self) -> bytes:
        schema = get_capnp_schema(schema_file="tensor.capnp")
        tensor_struct: CapnpModule = schema.Tensor  # type: ignore
        tensor_msg = tensor_struct.new_message()

        # this is how we dispatch correct deserialization of bytes
        tensor_msg.magicHeader = serde_magic_header(type(self))

        chunk_bytes(sy.serialize(self.child, to_bytes=True), "child", tensor_msg)

        tensor_msg.publicShape = sy.serialize(self.public_shape, to_bytes=True)

        # upcast the String class before setting to capnp
        public_dtype_func = getattr(
            self.public_dtype, "upcast", lambda: self.public_dtype
        )
        tensor_msg.publicDtype = public_dtype_func()
        tensor_msg.tagName = self.tag_name

        return tensor_msg.to_bytes_packed()

    @staticmethod
    def _bytes2object(buf: bytes) -> Tensor:
        schema = get_capnp_schema(schema_file="tensor.capnp")
        tensor_struct: CapnpModule = schema.Tensor  # type: ignore
        # https://stackoverflow.com/questions/48458839/capnproto-maximum-filesize
        MAX_TRAVERSAL_LIMIT = 2**64 - 1
        tensor_msg = tensor_struct.from_bytes_packed(
            buf, traversal_limit_in_words=MAX_TRAVERSAL_LIMIT
        )

        tensor = Tensor(
            child=sy.deserialize(combine_bytes(tensor_msg.child), from_bytes=True),
            public_shape=sy.deserialize(tensor_msg.publicShape, from_bytes=True),
            public_dtype=tensor_msg.publicDtype,
        )
        tensor.tag_name = tensor_msg.tagName

        return tensor



# future
from __future__ import annotations

# stdlib
from typing import Any
from typing import Dict
from typing import Optional

# third party
from nacl.signing import VerifyKey

# relative
from .....common.message import ImmediateSyftMessageWithReply
from .....common.message import ImmediateSyftMessageWithoutReply
from .....common.uid import UID
from .....io.address import Address
from ....abstract.node_service_interface import NodeServiceInterface


class GenericPayloadMessage(ImmediateSyftMessageWithReply):
    __attr_allowlist__ = ["id", "payload", "address", "reply_to", "msg_id"]

    def __init__(
        self,
        payload: GenericPayloadMessageWithReply,
        address: Address,
        reply_to: Address,
        msg_id: Optional[UID] = None,
    ):
        super().__init__(address=address, msg_id=msg_id, reply_to=reply_to)
        self.payload = payload


class GenericPayloadReplyMessage(ImmediateSyftMessageWithoutReply):
    __attr_allowlist__ = ["id", "payload", "address", "msg_id"]

    def __init__(
        self,
        payload: GenericPayloadMessageWithReply,
        address: Address,
        msg_id: Optional[UID] = None,
    ):
        super().__init__(address=address, msg_id=msg_id)
        self.payload = payload


class GenericPayloadMessageWithReply:
    __attr_allowlist__ = ["kwargs"]
    message_type = GenericPayloadMessage
    message_reply_type = GenericPayloadReplyMessage

    def __init__(self, kwargs: Optional[Dict[str, Any]] = None) -> None:
        if hasattr(kwargs, "upcast"):
            kwargs = kwargs.upcast()  # type: ignore
        self.kwargs: Dict[str, Any] = kwargs if kwargs is not None else {}

    def run(
        self, node: NodeServiceInterface, verify_key: Optional[VerifyKey] = None
    ) -> Dict[str, Any]:
        raise NotImplementedError

    def to(self, address: Address, reply_to: Address) -> Any:
        return self.message_type(address=address, reply_to=reply_to, payload=self)

    def back_to(self, address: Address) -> Any:
        return self.message_reply_type(address=address, payload=self)
